unit ICComponent;

(* Start Header *)

(* File:         ICComponent.p
 * Generated by: 1.0d5
 * For:          IC 1.4
 * On:           Saturday, 22 February 1997, 17:54:20
 * 
 * This file is part of the Internet Configuration system and
 * is placed in the public domain for the benefit of all.
 *)

(* End Header *)

interface

	uses
		Components;

{$main}
	function Main (var params: ComponentParameters; storage: Handle): ComponentResult;

implementation

	uses
		Types,
		Files,
		Memory, 
		Errors, 
		ToolUtils, 
		AppleTalk, 
		Aliases, 
		Packages, 
		GestaltEqu, 
		Resources, 
		Dialogs,
		AppleEvents, 
		Folders, 

		{$ifc not undefined MWERKS}
			PascalA4,
		{$endc}

		ICTypes, 
		ICKeys, 

		ICResourceForkSanity, 
		ICLinkInSubs, 
		ICLinkIn, 
		ICCAPI, 
		ICComponentSelectors;

	(* Component Globals
	
		The component maintains very few globals variables.  The most important
		global is "inst", a instance of the ICRRecord.  See the comment in ICLinkIn.p
		for more information about the ICRRecord.
		
		Note that the component uses the "instance" field of ICRRecord as a
		"target" variable, ie the place it maintains the component's current target.
		
		The other fields are a "self" reference (mostly used to pass back to
		various Component Manager routines that require a reference to
		the component) and a local copy of the componentVersion.
	*)

	type
		ComponentGlobals =
			record
				self: ComponentInstance;
				haveStarted : Boolean;
				inst: ICRRecord;
				componentVersion: NumVersion;
				oldResourceFileRefNum : integer;
				resourceFileRefNum : integer;
			end;
		ComponentGlobalsPtr = ^ComponentGlobals;
		ComponentGlobalsHandle = ^ComponentGlobalsPtr;

	(* ***** Component Manager Routines ***** *)
	
	(* Routines that are generally implemented by components. *)

	function ICCICanDo (globals: ComponentGlobalsHandle; selector: integer): ComponentResult;
		(* Handle the Component Manager CanDo request. We can do a standard set
			of Component Manager defined selectors, and all the selectors defined
			in ICComponentSelectors.p.
		*)
	begin
		{$unused globals}
		case selector of
			kComponentVersionSelect..kComponentOpenSelect,
				kComponentTargetSelect,
					kICCFirstSelector..kICCLastSelector: 
				ICCICanDo := 1;
			otherwise
				ICCICanDo := 0;
		end; (* case *)
	end; (* ICCICanDo *)

	function ICCIOpen (globals: ComponentGlobalsHandle; self: ComponentInstance): ComponentResult;
		(* Handle the Component Manager Open request.  We do very little here, basically
			using it as an opportunity to create our globals.  Most of the real initialisation
			is done in the Start routine.
		*)
		var
			err: ComponentResult;
	begin
		(* Create our globals. *)
		globals := ComponentGlobalsHandle(NewHandleClear(sizeof(ComponentGlobals)));
		err := MemError;

		(* Initialise some specified fields and tell the Component Manager about our globals. *)
		if err = noErr then begin
			globals^^.self := self;
			globals^^.inst.instance := self;
			SetComponentInstanceStorage(self, Handle(globals));
		end; (* if *)

		ICCIOpen := noErr;
	end; (* ICCIOpen *)

	function ICCIClose (globals: ComponentGlobalsHandle; self: ComponentInstance): ComponentResult;
		(* Handle the Component Manager Close request.  We presume that the client
			has correctly called Stop, so all we have to do is dispose of our globals
			and leave.  At the moment the component doesn't shut down cleanly
			if you call Close without calling Stop.  I may address that in the future,
			but at the moment the only interesting state we have is a possibly
			open the resource file, which will be cleaned up on application termination
			anyway.
		*)
		var
			err: ComponentResult;
			junk : ICError;
	begin
		{$unused self}
		err := noErr;
		if globals <> nil then begin
			if globals^^.haveStarted then begin
				junk := ICRStop(globals^^.inst);
				globals^^.haveStarted := false;
			end; (* if *)
			DisposeHandle(Handle(globals));
		end; (* if *)
		ICCIClose := err;
	end; (* ICCIClose *)

	function ICCITarget (globals: ComponentGlobalsHandle; newTarget: ComponentInstance): ComponentResult;
		(* Handle the Component Manager Target.  We change our current target
			variable to reflect the new target.
		*)
	begin
		globals^^.inst.instance := newTarget;
		ICCITarget := noErr;
	end; (* ICCITarget *)

	(* ***** Resources and Resource Management ***** *)

	const
		(* STR# *)
		rMiscStrings = 1024;
			strDamaged = 1;
			strDamagedBeyondRepair = 2;
			strBackupInTrash = 3;
			strDamagedSuffix = 4;
			strDefaultFileName = 5;
			strNewConfigPrompt = 6;

		(* ALRT *)
		rDamagedAlert = 1024;
	
	(* The component regularly needs access to its own resources.
		In some cases, subroutines need access to resources but they don't
		know whether the calling routine has already opened the resource
		file.  [The Component Manager's OpenComponentResFile routine
		makes no attempt to reference count the resource file.]  So we
		build our own reference counting mechanims.
		
		You can use the following two routines in two different ways:
		
			err := StartUsingComponentResources(didOpen);
			if err = noErr begin
				// do stuff
				StopUsingComponentResources(didOpen)
			end
		
		or
		
			err := StartUsingComponentResources(didOpen);
			if err = noErr begin
				// do stuff
			end
			StopUsingComponentResources(didOpen)
		
		Both of these should work just hunky dory.
	*)

	function StartUsingComponentResources(globals: ComponentGlobalsHandle; var didOpen : Boolean) : ICError;
		(* Makes the component's resources available.  Sets didOpen to true if it actually
			opened the resource file.  If this returns noErr, you should make sure you make
			a matching call to StopUsingComponentResources.
		*)
		var
			err : ICError;
	begin
		err := noErr;
		if globals^^.resourceFileRefNum <> 0 then begin
			didOpen := false;
		end else begin
			globals^^.oldResourceFileRefNum := CurResFile;
			globals^^.resourceFileRefNum := OpenComponentResFile(Component(globals^^.self));
			if globals^^.resourceFileRefNum <= 0 then begin
				globals^^.resourceFileRefNum := 0;
				err := resNotFound;
			end; (* if *)
			didOpen := (err = noErr);
		end; (* if *)
		StartUsingComponentResources := err;
	end; (* StartUsingComponentResources *)

	function StopUsingComponentResources(globals: ComponentGlobalsHandle; didOpen : Boolean) : ICError;
		(* Closes the component's resource file if we opened it. *)
		var
			err : ICError;
	begin
		err := noErr;
		if didOpen then begin
			err := CloseComponentResFile(globals^^.resourceFileRefNum);
			if err = noErr then begin
				globals^^.resourceFileRefNum := 0;
				UseResFile(globals^^.oldResourceFileRefNum);
			end; (* if *)
		end; (* if *)
		StopUsingComponentResources := err;
	end; (* StopUsingComponentResources *)

	function GetComponentIndString (globals: ComponentGlobalsHandle; 
																strListID: integer; index: integer; str: StringPtr): ICError;
		(* Get a string from a STR# resource out of the component resource fork.
			One important feature of this routine is that it doesn't modify str unless
			it succeeds in getting a non-empty string.  This is to assist clients
			that go:
				str := 'hardwire constant';
				err := GetComponentIndString(...);
			and junk the error code, assuming that the hardwired constant is
			good enough in the presence of errors.
		*)
		var
			err: ICError;
			junk: ICError;
			didOpen : Boolean;
			tmpStr : Str255;
	begin
		err := StartUsingComponentResources(globals, didOpen);
		if err = noErr then begin
			GetIndString(tmpStr, strListID, index);
			if tmpStr <> '' then begin
				(* Only copy out the required number of bytes, ie the number
					of bytes that are actually in the string. This is because
					lots of our clients pass in things that aren't Str255's
					(ie Str31s and Str63s), and we don't want to blat
					beyond what they've allocated.
				*)
				BlockMoveData(@tmpStr[0], str, length(tmpStr) + 1);
			end; (* if *)
			junk := StopUsingComponentResources(globals, didOpen);
		end; (* if *)
		GetComponentIndString := err;
	end; (* GetComponentIndString *)

	(* ***** Utility Routines ***** *)

	type
		IntegerHandle = ^IntegerPtr;

	function ValidStrList (strListH: Handle): Boolean;
		(* This function returns true if strListH is a valid string list. *)
		var
			result: Boolean;
			count: integer;
			position: longint;
			strListSize: longint;
			numberOfStringsFound: integer;
	begin
		result := (strListH <> nil) & (strListH^ <> nil) & (GetHandleSize(strListH) >= sizeof(integer));
		if result then begin
			count := IntegerPtr(strListH^)^;
			position := sizeof(integer);
			strListSize := GetHandleSize(strListH);
			numberOfStringsFound := 0;

			while result & (position < strListSize) do begin
				position := position + band(Ptr(longint(strListH^) + position)^, $00FF) + 1;
				if position > strListSize then begin
					result := false;
				end else begin
					numberOfStringsFound := numberOfStringsFound + 1;
				end; (* if *)
			end; (* while *)

			if result & ((numberOfStringsFound <> count) | (position <> strListSize)) then begin
				result := false;
			end; (* if *)
		end; (* if *)
		ValidStrList := result;
	end; (* ValidStrList *)

	function GetIndStrList (strListH: Handle; index: integer): Str255;
		(* This function returns the index'th string in the string list. If index
			is out of bounds, it returns the empty string.
		*)
		var
			result: Str255;
			count: integer;
			stringListIndex: integer;
			position: longint;
	begin
		count := IntegerHandle(strListH)^^;
		if (1 <= index) and (index <= count) then begin
			position := sizeof(integer);
			for stringListIndex := 1 to index - 1 do begin
				position := position + band(Ptr(ord(strListH^) + position)^, $FF) + 1;
			end; (* for *)
			BlockMoveData(Ptr(ord(strListH^) + position), @result, band(Ptr(ord(strListH^) + position)^, $FF) + 1);
		end else begin
			result := '';
		end; (* if *)
		GetIndStrList := result;
	end; (* GetIndStrList *)

	function CopyFork (sourceForkRefnum, destForkRefnum: integer; bytesToCopy: longint): OSErr;
		(* This routine copies a file fork from sourceForkRefnum to destForkRefnum.
			The files must be positioned at where you want to start copying (usually
			at the beginning) and the routine copys bytesToCopy bytes from the source
			to the destination.
		*)
		const
			kMaxCopyBufferSize = 65536;
			kMinCopyBufferSize = 512;
		var
			err: OSErr;
			copyBuffer: Ptr;
			copyBufferSize: longint;
			numberOfBytesThisTime: longint;
	begin
		err := noErr;
		copyBufferSize := kMaxCopyBufferSize;
		copyBuffer := nil;
		repeat
			copyBuffer := NewPtr(copyBufferSize);
			if copyBuffer = nil then begin
				copyBufferSize := copyBufferSize div 2;
			end; (* if *)
		until (copyBuffer <> nil) or (copyBufferSize < kMinCopyBufferSize);
		if copyBuffer = nil then begin
			err := memFullErr;
		end; (* if *)

		(* Now copy the file data, in copyBufferSize chunks. *)
		while (err = noErr) & (bytesToCopy > 0) do begin
			numberOfBytesThisTime := copyBufferSize;
			if numberOfBytesThisTime > bytesToCopy then begin
				numberOfBytesThisTime := bytesToCopy;
			end; (* if *)
			err := FSRead(sourceForkRefnum, numberOfBytesThisTime, copyBuffer);
			if err = noErr then begin
				bytesToCopy := bytesToCopy - numberOfBytesThisTime;
				err := FSWrite(destForkRefnum, numberOfBytesThisTime, copyBuffer);
			end; (* if *)
		end; (* while *)

		(* Clean up. *)
		if copyBuffer <> nil then begin
			DisposePtr(copyBuffer);
		end; (* if *)
		CopyFork := err;
	end; (* CopyFork *)

	function CopyForkToFork (var sourceFile: FSSpec; var destFile: FSSpec;
													sourceRsrc: Boolean; destRsrc: Boolean): OSErr;
		(* This routine copies a fork from the sourceFile to the destFile.
			The fork chosen in each case is determined by the sourceRsrc and
			destRsrc switches.  If true, the resource fork is used, if false,
			the data fork.
		*)
		var
			err: OSErr;
			srcRefNum: integer;
			destRefNum: integer;
			sizeofSrcFork: longint;
			junk: OSErr;
	begin
		(* Prepare for failure. *)
		srcRefNum := 0;
		destRefNum := 0;
		
		(* Open the source fork. *)
		if sourceRsrc then begin
			err := FSpOpenRF(sourceFile, fsRdPerm, srcRefNum);
		end else begin
			err := FSpOpenDF(sourceFile, fsRdPerm, srcRefNum);
		end; (* if *)
		if err <> noErr then begin
			srcRefNum := 0;
		end; (* if *)
		
		(* Open the dest fork. *)
		if err = noErr then begin
			if destRsrc then begin
				err := FSpOpenRF(destFile, fsRdWrPerm, destRefNum);
			end else begin
				err := FSpOpenDF(destFile, fsRdWrPerm, destRefNum);
			end; (* if *)
			if err <> noErr then begin
				destRefNum := 0;
			end; (* if *)
		end; (* if *)

		(* Set the length of the dest fork to the length of the source fork. *)
		if err = noErr then begin
			err := GetEOF(srcRefNum, sizeofSrcFork);
		end; (* if *)
		if err = noErr then begin
			err := SetEOF(destRefNum, sizeofSrcFork);
		end; (* if *)

		(* Copy the fork. *)
		if err = noErr then begin
			err := CopyFork(srcRefNum, destRefNum, sizeofSrcFork);
		end; (* if *)

		(* Clean up. *)
		if srcRefNum <> 0 then begin
			junk := FSClose(srcRefNum);
		end; (* if *)
		if destRefNum <> 0 then begin
			junk := FSClose(destRefNum);
		end; (* if *)

		junk := FlushVol(nil, destFile.vRefNum);

		CopyForkToFork := err;
	end; (* CopyForkToFork *)

	function CopyFileForkToFork (var fss: FSSpec; rsrcToData: Boolean): OSErr;
		(* This routine is a simple wrapper around CopyForkToFork.  It is
			used to make backups of the resource fork in the data fork,
			and to unbackup from the data fork to the resource fork.
			fss is the file to backup.  If rsrcToData is true, it copies the resource
			fork to the data fork, otherwise vice versa.
		*)
	begin
		CopyFileForkToFork := CopyForkToFork(fss, fss, rsrcToData, not rsrcToData);
	end; (* CopyFileForkToFork *)

	function BackupFileIntoTrash (globals: ComponentGlobalsHandle): OSErr;
		(* Make a copy of the preferences file in the Trash.  This function
			is called when IC detects that a preferences file is corrupt.
			Instead of just overwriting the file, it copies it to the Trash
			and then overwrites it.  This way, the user (or, more importantly, I)
			can analyse the file to see what symptoms it displays.
		*)
		var
			err: OSErr;
			junk: OSErr;
			createdBackupFile: Boolean;
			damagedFileCatalogInfo: CInfoPBRec;
			backupFile: FSSpec;
			tries: integer;
			middlePartString: Str31;
			triesString: Str255;
	begin
		(* Prepare for failure.  We remember whether we created a backup
			file so that we can delete it on failure at the end of this routine.
		*)
		createdBackupFile := false;

		(* Fail if we don't have FindFolder.  We only do these smarts under
			System 7.  The reason for this is that creating a file in the trash
			is tricky under System 6 -- you have to mess around with Finder
			flags and so on.
		*)
		err := noErr;
		if not ICUHaveFindFolder then begin
			err := unimpErr;
		end; (* if *)

		(* Get info about the damaged file (ie the currently configured file) to
			use when we create the copy.
		*)
		if err = noErr then begin
			err := ICUFSpGetCatInfo(globals^^.inst.configFileSpec, 0, damagedFileCatalogInfo);
		end; (* if *)

		(* Now create an FSSpec that points to the file we'd like to create in the Trash. *)
		if err = noErr then begin
			backupFile.name := globals^^.inst.configFileSpec.name;
			err := ICUFindFolder(kOnSystemDisk, kTrashFolderType, kCreateFolder, backupFile.vRefNum, backupFile.parID);
		end; (* if *)

		(* Try and create the file in the Trash.  This involves creating it and, if we fail,
			adding an incrementing numeric index to the end of the name.  We do this
			up to 100 times before bailing out completely.
		*)
		if err = noErr then begin
			middlePartString := ' (Damaged)';
			junk := GetComponentIndString(globals, rMiscStrings, strDamagedSuffix, @middlePartString);

			tries := 0;
			triesString := '';
			repeat
				(* Do some magic to merge the various file names.  Basically we have
					3 components that we'd like to concatenate:
						1. the damaged file name
						2. the middle part, ie " (Damaged)"
						3. the tries string, ie "" the first time, " 1" the second time, and so on.
					We first try concating them to see if they'll all find into 31 characters.
					If not, we reduce the length of the first component (the damaged file
					name) until the name fits into 31 characters.
				*)
				if length(concat(globals^^.inst.configFileSpec.name,
											middlePartString, 
											triesString)) > 31 then begin
					backupFile.name := concat(copy(globals^^.inst.configFileSpec.name,
																		1, 
																		31 - length(middlePartString) - length(triesString)), 
																middlePartString, 
																triesString);
				end else begin
					backupFile.name := concat(globals^^.inst.configFileSpec.name, 
																middlePartString, 
																triesString);
				end; (* if *)
				err := HCreate(backupFile.vRefNum, backupFile.parID, backupFile.name, damagedFileCatalogInfo.ioFlFndrInfo.fdCreator, damagedFileCatalogInfo.ioFlFndrInfo.fdType);
				tries := tries + 1;
				NumToString(tries, triesString);
				triesString := concat(' ', triesString);
			until (err = noErr) or (tries > 100);

		end; (* if *)

		(* Copy the forks to the newly created file, then set it's Finder info
			appropriately.
		*)
		if err = noErr then begin
			createdBackupFile := true;
			err := CopyForkToFork(globals^^.inst.configFileSpec, backupFile, false, false);
		end; (* if *)
		if err = noErr then begin
			err := CopyForkToFork(globals^^.inst.configFileSpec, backupFile, true, true);
		end; (* if *)
		if err = noErr then begin
			(* Clear the "inited" bit. *)
			damagedFileCatalogInfo.ioFlFndrInfo.fdFlags := band(damagedFileCatalogInfo.ioFlFndrInfo.fdFlags, $FFFFFEFF);
			err := ICUFSpSetCatInfo(backupFile, damagedFileCatalogInfo);
		end; (* if *)

		(* Clean up. *)
		if (err <> noErr) and createdBackupFile then begin
			junk := HDelete(backupFile.vRefNum, backupFile.parID, backupFile.name);
		end; (* if *)

		BackupFileIntoTrash := err;
	end; (* BackupFileIntoTrash *)
	
	procedure DisplayFileFateDialog (globals: ComponentGlobalsHandle;
															irrecoverably: Boolean; 
															backupMade: Boolean);
		(* Display a dialog showing the user the fate of their Internet Preferences file.
			If irrecoverably is set, say that the file was reverted to defaults, otherwise
			say it was merely reverted to a backup.  If backupMade is set, say the damaged
			copy is in the Trash, otherwise say nothing.
		*)
		var
			err: OSErr;
			junk: integer;
			backupMadeMessage: Str255;
			damageReport : Str255;
			didOpen : Boolean;
	begin
		err := ICUCanInteract;
		if err = noErr then begin
			err := StartUsingComponentResources(globals, didOpen);
		end; (* if *)
		if err = noErr then begin
			if backupMade then begin
				junk := GetComponentIndString(globals, rMiscStrings, strBackupInTrash, @backupMadeMessage);
			end else begin
				backupMadeMessage := '';
			end; (* if *)
			damageReport := '';
			junk := GetComponentIndString(globals, rMiscStrings, strDamaged + ord(irrecoverably), @damageReport);
			ParamText(globals^^.inst.configFileSpec.name, damageReport, backupMadeMessage, '');
			junk := NoteAlert(rDamagedAlert, nil);
			
			junk := StopUsingComponentResources(globals, didOpen);
		end; (* if *)
	end; (* DisplayFileFateDialog *)

	(* ***** Default Preference Creation ***** *)

	function GetOwnerName: Str255;
		(* Returns the Owner Name, as configured in the Sharing Setup control panel. *)
		const
			rOwnerNameString = -16096;
		var
			strH: StringHandle;
	begin
		strH := GetString(rOwnerNameString);
		if strH <> nil then begin
			(* Don't release it, someone else might be using it. *)
			GetOwnerName := strH^^;
		end else begin
			GetOwnerName := '';
		end; (* if *)
	end; (* GetOwnerName *)

	procedure AddComponentDefaultPrefs (globals: ComponentGlobalsHandle);
		(* Add the default preferences to the currently configured preferences
			file.  The basic algorithm is to copy across all the 'Pref' resources
			from component's resource fork to 'ICRP' resources in the preference
			file.  After that, we have to do some special case stuff, like
			selecting a default set of servers and configuring the RealName
			preference from the Owner Name.
		*)
		var
			prefFileResourceRefnum: integer;

		function PrefExists (key: Str255): Boolean;
			var
				prefH: Handle;
		begin
			UseResFile(prefFileResourceRefnum);
			SetResLoad(false);
			prefH := Get1NamedResource(kICPrefResourceType, key);
			SetResLoad(true);
			(* We don't ReleaseResource(prefH) because it will be cleaned up when we close*)
			(* the preferences file at the end of AddDefaultPrefs. *)
			PrefExists := (prefH <> nil);
		end; (* PrefExists *)

		procedure SetPrefHandle (key: Str255; var dataH: Handle);
			(* Store the data in dataH with the specified key in the current document. *)
			var
				err: OSErr;
				junkLong: longint;
				emptyAttr: ICAttr;
				id: integer;
		begin
			if not PrefExists(key) then begin
				emptyAttr := 0;
				junkLong := Munger(dataH, 0, nil, 0, @emptyAttr, sizeof(emptyAttr));
				err := MemError;
				if err = noErr then begin
					repeat
						id := Unique1ID(kICPrefResourceType);
					until id > 127;
					AddResource(dataH, kICPrefResourceType, id, key);
					err := ResError;
				end; (* if *)
				if err = noErr then begin
					dataH := nil;
				end; (* if *)
			end; (* if *)
		end; (* SetPrefHandle *)

		procedure SetPrefStr (key: Str255; data: Str255);
			(* Stores the data in the data string with the specified key in current document. *)
			var
				stringH: StringHandle;
		begin
			if not PrefExists(key) then begin
				stringH := NewString(data);
				if stringH <> nil then begin
					SetPrefHandle(key, Handle(stringH));
				end; (* if *)
				if stringH <> nil then begin
					DisposeHandle(Handle(stringH));
				end; (* if *)
			end; (* if *)
		end; (* SetPrefStr *)

		procedure CheckServerDefaultPref (key: Str255);
			var
				prefH: Handle;
				junk: longint;
		begin
			if not PrefExists(concat(key, 'Preferred')) then begin
				prefH := Get1NamedResource(kICPrefResourceType, concat(key, 'All'));
				if prefH <> nil then begin

					(* I use HandToHand to copy the resource rather than detaching it because *)
					(* the resource has not yet been written to the resource file yet, so *)
					(* it's not possible to detach it.  DetachResource returns an error. *)
					(* So instead I clone the handle and proceed from there. *)

					if HandToHand(prefH) = noErr then begin
						junk := Munger(prefH, 0, nil, sizeof(ICAttr), Ptr(-1), 0);			(* Strip off leading attributes. *)
						if ValidStrList(prefH) then begin
							SetPrefStr(concat(key, 'Preferred'), GetIndStrList(prefH, 1));
						end; (* if *)
						DisposeHandle(prefH);
					end; (* if *)
				end; (* if *)
			end; (* if *)
		end; (* CheckServerDefaultPref *)

		procedure CopyAcrossPrefResources;
			(* Copies all the 'Pref' resources from the component
				resource fork to the preferences file resource fork.
			*)
			var
				numberOfPrefResources: integer;
				prefResourceIndex : integer;
				prefData: Handle;
				prefName: Str255;
				junkID: integer;
				junkType: ResType;
		begin
			(* First copy across the 'Pref' resources. *)
			UseResFile(globals^^.resourceFileRefNum);
			numberOfPrefResources := Count1Resources('Pref');
			for prefResourceIndex := 1 to numberOfPrefResources do begin
				UseResFile(globals^^.resourceFileRefNum);
				prefData := Get1IndResource('Pref', prefResourceIndex);
				if prefData <> nil then begin
					GetResInfo(prefData, junkID, junkType, prefName);
					DetachResource(prefData);
					SetPrefHandle(prefName, prefData);
					if prefData <> nil then begin
						DisposeHandle(prefData);
					end; (* if *)
				end; (* if *)
			end; (* if *)
		end; (* CopyAcrossPrefResources *)

		var
			err: OSErr;
			prefFileCreatorVersionH: VersRecHndl;
			componentVersionH: VersRecHndl;
			junk: OSErr;
			didOpen : Boolean;
	begin
		prefFileResourceRefnum := HOpenResFile(globals^^.inst.configFileSpec.vRefNum, 
																			globals^^.inst.configFileSpec.parID, 
																				globals^^.inst.configFileSpec.name, 
																					fsRdWrPerm);
		err := ResError;
		if err = noErr then begin
			prefFileCreatorVersionH := VersRecHndl(Get1Resource('vers', 1));
			if (prefFileCreatorVersionH = nil) | (longint(prefFileCreatorVersionH^^.numericVersion) < longint(globals^^.componentVersion)) then begin

				err := StartUsingComponentResources(globals, didOpen);
				if err = noErr then begin
					CopyAcrossPrefResources;

					(* Then make sure that the servers have a valid default value. *)
					CheckServerDefaultPref('Archie');
					CheckServerDefaultPref('InfoMac');
					CheckServerDefaultPref('UMich');

					(* Set up the Real Name preference from the owner name resource *)
					(*	in the System file. *)
					SetPrefStr(kICRealName, GetOwnerName);

					(* Get the current component version and store it in the preferences file. *)
					UseResFile(globals^^.resourceFileRefNum);
					componentVersionH := VersRecHndl(Get1Resource('vers', 1));
					if componentVersionH <> nil then begin
						DetachResource(Handle(componentVersionH));
						UseResFile(prefFileResourceRefnum);
						if prefFileCreatorVersionH = nil then begin
							AddResource(Handle(componentVersionH), 'vers', 1, '');
							if ResError = noErr then begin
								componentVersionH := nil;
							end; (* if *)
						end else begin
							HLock(Handle(componentVersionH));
							junk := PtrToXHand(Ptr(componentVersionH^), Handle(prefFileCreatorVersionH), GetHandleSize(Handle(componentVersionH)));
							ChangedResource(Handle(prefFileCreatorVersionH));
						end; (* if *)
						if componentVersionH <> nil then begin
							DisposeHandle(Handle(componentVersionH));
						end; (* if *)
					end; (* if *)

					junk := StopUsingComponentResources(globals, didOpen);
				end; (* if *)
			end; (* if *)
			CloseResFile(prefFileResourceRefnum);
		end; (* if *)
	end; (* AddComponentDefaultPrefs *)

	(* ***** Component Specific Routines ***** *)
	
	(* These are the API routines that the IC component overrides, to add
		some value on top of the link-in implementation that most people
		use.
	*)

	function ICCIStart (globals: ComponentGlobalsHandle; creator: OSType): ICError;
		(* Handle the start request.  This is basically a replacement for the Open
			because we need bonus data (creator).  This implementation calls though
			to the link-in implementation and then grabs some resources that
			we need to have conveniently at hand.
		*)
		var
			err: OSErr;
			didOpen : Boolean;
			tmpH: Handle;
			junk: OSErr;
			tmpStr : Str255;
	begin
		err := ICRStart(globals^^.inst, creator);
		if err = noErr then begin
			globals^^.haveStarted := true;
			err := StartUsingComponentResources(globals, didOpen);
			if err = noErr then begin
				
				(* Setup the prompt string. *)
				if GetComponentIndString (globals,  rMiscStrings, strNewConfigPrompt, @tmpStr) = noErr then begin
					globals^^.inst.promptString := tmpStr;
				end else begin
					(* On error, leave the prompt as it was set up by ICRStart. *)
				end; (* if *)

				(* Setup our global knowledge about the component's version number. *)
				tmpH := Get1Resource('vers', 1);
				if tmpH = nil then begin
					err := resNotFound;
				end else begin
					globals^^.componentVersion := VersRecHndl(tmpH)^^.numericVersion;
				end; (* if *)

				junk := StopUsingComponentResources(globals, didOpen);
			end; (* if *)
		end; (* if *)
		ICCIStart := err;
	end; (* ICCIStart *)

	function ICCIDefaultFileName (globals: ComponentGlobalsHandle; var name: Str63): ICError;
		(* Gets the default file name for Internet Preferences files.  The component
			version of this routine gets that name out of a resource.
		*)
	begin
		if GetComponentIndString(globals, rMiscStrings, strDefaultFileName, @name) <> noErr then begin
			name := ICdefault_file_name;
		end; (* if *)
		ICCIDefaultFileName := noErr;
	end; (* ICCIDefaultFileName *)

	function ICCIStop (globals: ComponentGlobalsHandle): ICError;
		(* The component version of the Stop routine detects whether we're
			stopping an instance that points to a modified file.  If we are,
			it backs up the resource fork in the file's data fork.
		*)
		var
			junk: OSErr;
	begin
		if globals^^.inst.thisFileModified then begin
			junk := CopyFileForkToFork(globals^^.inst.configFileSpec, true);
		end; (* if *)
		ICCIStop := ICRStop(globals^^.inst);
	end; (* ICCIStop *)

	function ICCISpecifyConfigFile (globals: ComponentGlobalsHandle; var config: FSSpec): ICError;
		(* The component version of the SpecifyConfigFile routine detects
			whether we're pointing the instance away from a modified file.
			If we are, it backs up the resource fork in the file's data fork.
		*)
		var
			junk: OSErr;
	begin
		if globals^^.inst.thisFileModified then begin
			junk := CopyFileForkToFork(globals^^.inst.configFileSpec, true);

			(* Note that ICRSpecifyConfigFile will clear thisFileModified. *)
		end; (* if *)
		ICCISpecifyConfigFile := ICRSpecifyConfigFile(globals^^.inst, config);
	end; (* ICCISpecifyConfigFile *)

	function ICCIBegin (globals: ComponentGlobalsHandle; perm: ICPerm): ICError;
		(* The component version of the Begin routine has special logic for
			detecting corrupt preferences file.  The logic goes:
			
				If it's the first time around, then invoke our special checks
					If the file is missing, we should create it
					if the file currently exists, we should validate it
						If the validation fails, we should
							Copy the file to the Trash
							Revert to backup
						If the reverted file still won't validate, we delete the entire thing
						Either way, we need to tell the user of the file's fate
					To finish up, we should AddComponentDefaultPrefs.
		*)
		var
			err: OSErr;
			junk: OSErr;
			sane: Boolean;
			cpb: CInfoPBRec;
			fileFate: (ffNone, ffDamagedAndReverted, ffDamagedBeyondRepair);
			backupMade: Boolean;
			refNum: integer;
	begin
		(* Only invoke the special logic the first time this instance
			calls Begin.
		*)
		err := noErr;
		if globals^^.inst.waitingForFirstBegin then begin
			fileFate := ffNone;

			(* Check whether the file exists and has a resource fork.  If it fails,
				go create the file.
			*)
			err := ICUFSpGetCatInfo (globals^^.inst.configFileSpec, 0, cpb);
			if (err = fnfErr) | ((err = noErr) & (cpb.ioFlRLgLen = 0)) then begin
				junk := HCreate(globals^^.inst.configFileSpec.vRefNum, 
										globals^^.inst.configFileSpec.parID, 
										globals^^.inst.configFileSpec.name, 
										ICcreator, ICfiletype);
				HCreateResFile(globals^^.inst.configFileSpec.vRefNum, 
											globals^^.inst.configFileSpec.parID, 
											globals^^.inst.configFileSpec.name);
			end; (* if *)

			(* Check the file for sanity.  If it fails, try to back it up
				and then revert it to the the backup in the data fork.
			*)
			err := noErr;
			sane := false;
			junk := CheckResourceForkSanity(globals^^.inst.configFileSpec, sane);
			if not sane then begin
				backupMade := (BackupFileIntoTrash(globals) = noErr);
				fileFate := ffDamagedAndReverted;
				err := CopyFileForkToFork(globals^^.inst.configFileSpec, false);
				if err = noErr then begin
					junk := CheckResourceForkSanity(globals^^.inst.configFileSpec, sane);
				end; (* if *)
			end; (* if *)

			(* Check the file for sanity again.  If it still fails, zap the file completely. *)
			if (err = noErr) & not sane then begin
				fileFate := ffDamagedBeyondRepair;
				err := HOpenRF(globals^^.inst.configFileSpec.vRefNum, globals^^.inst.configFileSpec.parID, globals^^.inst.configFileSpec.name, fsRdWrPerm, refNum);
				if err = noErr then begin
					err := SetEOF(refNum, 0);
					junk := FSClose(refNum);
				end; (* if *)

				(* Check the file for sanity one more time, just to make sure we're
					not completely in the weeds.
				*)
				if err = noErr then begin
					junk := CheckResourceForkSanity(globals^^.inst.configFileSpec, sane);
					if not sane then begin
						err := icInternalErr;
					end; (* if *)
				end; (* if *)
			end; (* if *)

			(* Finally, let the user know what we did to their preferences, and
				make sure the file has the default preferences in it.
			*)
			if err = noErr then begin
				if fileFate <> ffNone then begin
					DisplayFileFateDialog(globals, fileFate = ffDamagedBeyondRepair, backupMade);
				end; (* if *)
				AddComponentDefaultPrefs(globals);
			end; (* if *)
		end; (* if *)

		(* If all the above works, call through to the link-in implementation. *)
		if err = noErr then begin
			err := ICRBegin(globals^^.inst, perm);
		end; (* if *)
		ICCIBegin := err;
	end; (* ICCIBegin *)

	(* ***** Easy Component Routines ***** *)
	
	(* These routines all call straight through to their link-in implementation
		equivalents.  These versions are all automatically generated.
	*)
	
(* Start ICCI.p *)

  function ICCIFindConfigFile(globals : ComponentGlobalsHandle; count : integer; folders : ICDirSpecArrayPtr) : ICError;
  begin
    ICCIFindConfigFile := ICRFindConfigFile(globals^^.inst, count, folders);
  end; (* ICCIFindConfigFile *)
  
  function ICCIFindUserConfigFile(globals : ComponentGlobalsHandle; var where : ICDirSpec) : ICError;
  begin
    ICCIFindUserConfigFile := ICRFindUserConfigFile(globals^^.inst, where);
  end; (* ICCIFindUserConfigFile *)
  
  function ICCIGeneralFindConfigFile(globals : ComponentGlobalsHandle; searchPrefs : Boolean; canCreate : Boolean; count : integer; folders : ICDirSpecArrayPtr) : ICError;
  begin
    ICCIGeneralFindConfigFile := ICRGeneralFindConfigFile(globals^^.inst, searchPrefs, canCreate, count, folders);
  end; (* ICCIGeneralFindConfigFile *)
  
  function ICCIChooseConfig(globals : ComponentGlobalsHandle) : ICError;
  begin
    ICCIChooseConfig := ICRChooseConfig(globals^^.inst);
  end; (* ICCIChooseConfig *)
  
  function ICCIChooseNewConfig(globals : ComponentGlobalsHandle) : ICError;
  begin
    ICCIChooseNewConfig := ICRChooseNewConfig(globals^^.inst);
  end; (* ICCIChooseNewConfig *)
  
  function ICCIGetConfigName(globals : ComponentGlobalsHandle; longname : Boolean; var name : Str255) : ICError;
  begin
    ICCIGetConfigName := ICRGetConfigName(globals^^.inst, longname, name);
  end; (* ICCIGetConfigName *)
  
  function ICCIGetConfigReference(globals : ComponentGlobalsHandle; ref : ICConfigRefHandle) : ICError;
  begin
    ICCIGetConfigReference := ICRGetConfigReference(globals^^.inst, ref);
  end; (* ICCIGetConfigReference *)
  
  function ICCISetConfigReference(globals : ComponentGlobalsHandle; ref : ICConfigRefHandle; flags : longint) : ICError;
  begin
    ICCISetConfigReference := ICRSetConfigReference(globals^^.inst, ref, flags);
  end; (* ICCISetConfigReference *)
  
  function ICCIGetSeed(globals : ComponentGlobalsHandle; var seed : longint) : ICError;
  begin
    ICCIGetSeed := ICRGetSeed(globals^^.inst, seed);
  end; (* ICCIGetSeed *)
  
  function ICCIGetPerm(globals : ComponentGlobalsHandle; var perm : ICPerm) : ICError;
  begin
    ICCIGetPerm := ICRGetPerm(globals^^.inst, perm);
  end; (* ICCIGetPerm *)
  
  function ICCIGetPref(globals : ComponentGlobalsHandle; key : StringPtr; var attr : ICAttr; buf : Ptr; var size : longint) : ICError;
  begin
    ICCIGetPref := ICRGetPref(globals^^.inst, key, attr, buf, size);
  end; (* ICCIGetPref *)
  
  function ICCISetPref(globals : ComponentGlobalsHandle; key : StringPtr; attr : ICAttr; buf : Ptr; size : longint) : ICError;
  begin
    ICCISetPref := ICRSetPref(globals^^.inst, key, attr, buf, size);
  end; (* ICCISetPref *)
  
  function ICCIFindPrefHandle(globals : ComponentGlobalsHandle; key : StringPtr; var attr : ICAttr; prefh : Handle) : ICError;
  begin
    ICCIFindPrefHandle := ICRFindPrefHandle(globals^^.inst, key, attr, prefh);
  end; (* ICCIFindPrefHandle *)
  
  function ICCIGetPrefHandle(globals : ComponentGlobalsHandle; key : StringPtr; var attr : ICAttr; var prefh : Handle) : ICError;
  begin
    ICCIGetPrefHandle := ICRGetPrefHandle(globals^^.inst, key, attr, prefh);
  end; (* ICCIGetPrefHandle *)
  
  function ICCISetPrefHandle(globals : ComponentGlobalsHandle; key : StringPtr; attr : ICAttr; prefh : Handle) : ICError;
  begin
    ICCISetPrefHandle := ICRSetPrefHandle(globals^^.inst, key, attr, prefh);
  end; (* ICCISetPrefHandle *)
  
  function ICCICountPref(globals : ComponentGlobalsHandle; var count : longint) : ICError;
  begin
    ICCICountPref := ICRCountPref(globals^^.inst, count);
  end; (* ICCICountPref *)
  
  function ICCIGetIndPref(globals : ComponentGlobalsHandle; n : longint; var key : Str255) : ICError;
  begin
    ICCIGetIndPref := ICRGetIndPref(globals^^.inst, n, key);
  end; (* ICCIGetIndPref *)
  
  function ICCIDeletePref(globals : ComponentGlobalsHandle; key : StringPtr) : ICError;
  begin
    ICCIDeletePref := ICRDeletePref(globals^^.inst, key);
  end; (* ICCIDeletePref *)
  
  function ICCIEnd(globals : ComponentGlobalsHandle) : ICError;
  begin
    ICCIEnd := ICREnd(globals^^.inst);
  end; (* ICCIEnd *)
  
  function ICCIEditPreferences(globals : ComponentGlobalsHandle; key : StringPtr) : ICError;
  begin
    ICCIEditPreferences := ICREditPreferences(globals^^.inst, key);
  end; (* ICCIEditPreferences *)
  
  function ICCIParseURL(globals : ComponentGlobalsHandle; hint : StringPtr; data : Ptr; len : longint; var selStart : longint; var selEnd : longint; url : Handle) : ICError;
  begin
    ICCIParseURL := ICRParseURL(globals^^.inst, hint, data, len, selStart, selEnd, url);
  end; (* ICCIParseURL *)
  
  function ICCILaunchURL(globals : ComponentGlobalsHandle; hint : StringPtr; data : Ptr; len : longint; var selStart : longint; var selEnd : longint) : ICError;
  begin
    ICCILaunchURL := ICRLaunchURL(globals^^.inst, hint, data, len, selStart, selEnd);
  end; (* ICCILaunchURL *)
  
  function ICCIMapFilename(globals : ComponentGlobalsHandle; filename : StringPtr; var entry : ICMapEntry) : ICError;
  begin
    ICCIMapFilename := ICRMapFilename(globals^^.inst, filename, entry);
  end; (* ICCIMapFilename *)
  
  function ICCIMapTypeCreator(globals : ComponentGlobalsHandle; fType : OSType; fCreator : OSType; filename : StringPtr; var entry : ICMapEntry) : ICError;
  begin
    ICCIMapTypeCreator := ICRMapTypeCreator(globals^^.inst, fType, fCreator, filename, entry);
  end; (* ICCIMapTypeCreator *)
  
  function ICCIMapEntriesFilename(globals : ComponentGlobalsHandle; entries : Handle; filename : StringPtr; var entry : ICMapEntry) : ICError;
  begin
    ICCIMapEntriesFilename := ICRMapEntriesFilename(globals^^.inst, entries, filename, entry);
  end; (* ICCIMapEntriesFilename *)
  
  function ICCIMapEntriesTypeCreator(globals : ComponentGlobalsHandle; entries : Handle; fType : OSType; fCreator : OSType; filename : StringPtr; var entry : ICMapEntry) : ICError;
  begin
    ICCIMapEntriesTypeCreator := ICRMapEntriesTypeCreator(globals^^.inst, entries, fType, fCreator, filename, entry);
  end; (* ICCIMapEntriesTypeCreator *)
  
  function ICCICountMapEntries(globals : ComponentGlobalsHandle; entries : Handle; var count : longint) : ICError;
  begin
    ICCICountMapEntries := ICRCountMapEntries(globals^^.inst, entries, count);
  end; (* ICCICountMapEntries *)
  
  function ICCIGetIndMapEntry(globals : ComponentGlobalsHandle; entries : Handle; ndx : longint; var pos : longint; var entry : ICMapEntry) : ICError;
  begin
    ICCIGetIndMapEntry := ICRGetIndMapEntry(globals^^.inst, entries, ndx, pos, entry);
  end; (* ICCIGetIndMapEntry *)
  
  function ICCIGetMapEntry(globals : ComponentGlobalsHandle; entries : Handle; pos : longint; var entry : ICMapEntry) : ICError;
  begin
    ICCIGetMapEntry := ICRGetMapEntry(globals^^.inst, entries, pos, entry);
  end; (* ICCIGetMapEntry *)
  
  function ICCISetMapEntry(globals : ComponentGlobalsHandle; entries : Handle; pos : longint; var entry : ICMapEntry) : ICError;
  begin
    ICCISetMapEntry := ICRSetMapEntry(globals^^.inst, entries, pos, entry);
  end; (* ICCISetMapEntry *)
  
  function ICCIDeleteMapEntry(globals : ComponentGlobalsHandle; entries : Handle; pos : longint) : ICError;
  begin
    ICCIDeleteMapEntry := ICRDeleteMapEntry(globals^^.inst, entries, pos);
  end; (* ICCIDeleteMapEntry *)
  
  function ICCIAddMapEntry(globals : ComponentGlobalsHandle; entries : Handle; var entry : ICMapEntry) : ICError;
  begin
    ICCIAddMapEntry := ICRAddMapEntry(globals^^.inst, entries, entry);
  end; (* ICCIAddMapEntry *)
  

(* End ICCI.p *)

(* ----- Component Entry Point ----- *)

	function Main (var params: ComponentParameters; storage: Handle): ComponentResult;
		(* The main component entry point.  It's just a standard component dispatcher. *)
		var
			proc: ProcPtr;
			s: SignedByte;
			oldA4 : longint;
	begin
		{$ifc not undefined MWERKS} oldA4 := SetCurrentA4; {$endc}

		(* Setup proc based on which routine is to be called. *)
		proc := nil;
		case params.what of
			(* Component Manager stuff *)
			kComponentVersionSelect: 
				Main := internetConfigurationComponentInterfaceVersion + 2;
			kComponentCanDoSelect: 
				proc := @ICCICanDo;
			kComponentOpenSelect: 
				proc := @ICCIOpen;
			(* Component Manager stuff *)
			kComponentCloseSelect: 
				proc := @ICCIClose;
			kComponentTargetSelect: 
				proc := @ICCITarget;
			(* this component type stuff *)
(* Start ICCSel.p2 *)

        kICCStart: 
          proc := @ICCIStart;
        kICCStop: 
          proc := @ICCIStop;
        kICCFindConfigFile: 
          proc := @ICCIFindConfigFile;
        kICCFindUserConfigFile: 
          proc := @ICCIFindUserConfigFile;
        kICCGeneralFindConfigFile: 
          proc := @ICCIGeneralFindConfigFile;
        kICCChooseConfig: 
          proc := @ICCIChooseConfig;
        kICCChooseNewConfig: 
          proc := @ICCIChooseNewConfig;
        kICCGetConfigName: 
          proc := @ICCIGetConfigName;
        kICCGetConfigReference: 
          proc := @ICCIGetConfigReference;
        kICCSetConfigReference: 
          proc := @ICCISetConfigReference;
        kICCSpecifyConfigFile: 
          proc := @ICCISpecifyConfigFile;
        kICCGetSeed: 
          proc := @ICCIGetSeed;
        kICCGetPerm: 
          proc := @ICCIGetPerm;
        kICCDefaultFileName: 
          proc := @ICCIDefaultFileName;
        kICCBegin: 
          proc := @ICCIBegin;
        kICCGetPref: 
          proc := @ICCIGetPref;
        kICCSetPref: 
          proc := @ICCISetPref;
        kICCFindPrefHandle: 
          proc := @ICCIFindPrefHandle;
        kICCGetPrefHandle: 
          proc := @ICCIGetPrefHandle;
        kICCSetPrefHandle: 
          proc := @ICCISetPrefHandle;
        kICCCountPref: 
          proc := @ICCICountPref;
        kICCGetIndPref: 
          proc := @ICCIGetIndPref;
        kICCDeletePref: 
          proc := @ICCIDeletePref;
        kICCEnd: 
          proc := @ICCIEnd;
        kICCEditPreferences: 
          proc := @ICCIEditPreferences;
        kICCParseURL: 
          proc := @ICCIParseURL;
        kICCLaunchURL: 
          proc := @ICCILaunchURL;
        kICCMapFilename: 
          proc := @ICCIMapFilename;
        kICCMapTypeCreator: 
          proc := @ICCIMapTypeCreator;
        kICCMapEntriesFilename: 
          proc := @ICCIMapEntriesFilename;
        kICCMapEntriesTypeCreator: 
          proc := @ICCIMapEntriesTypeCreator;
        kICCCountMapEntries: 
          proc := @ICCICountMapEntries;
        kICCGetIndMapEntry: 
          proc := @ICCIGetIndMapEntry;
        kICCGetMapEntry: 
          proc := @ICCIGetMapEntry;
        kICCSetMapEntry: 
          proc := @ICCISetMapEntry;
        kICCDeleteMapEntry: 
          proc := @ICCIDeleteMapEntry;
        kICCAddMapEntry: 
          proc := @ICCIAddMapEntry;

(* End ICCSel.p2 *)
			otherwise
				Main := badComponentSelector;
		end; (* case *)

		(* If we've got a routine to call, call it.  Note that the default case
			above has set Main up to return the appropriate error code if
			proc is nil.
		*)
		if proc <> nil then begin
			(* Lock down the storage handle. *)
			if storage <> nil then begin
				s := HGetState(storage);
				HLock(storage);
			end; (* if *)
			
			(* Dispatch to the relevant routine. *)
			Main := CallComponentFunctionWithStorage(storage, params, proc);
			
			(* Unlock the storage handle, except in the Close case, where the
				handle has already been disposed.
			*)
			if (storage <> nil) and (params.what <> kComponentCloseSelect) then begin
				HSetState(storage, s);
			end; (* if *)
		end; (* if *)
		{$ifc not undefined MWERKS} oldA4 := SetA4(oldA4); {$endc}
	end; (* Main *)

end. (* ICComponent *)