unit ICLinkIn;

(* Start Header *)

(* File:         ICLinkIn.p
 * Generated by: 1.0d5
 * For:          IC 1.4
 * On:           Saturday, 22 February 1997, 17:54:20
 * 
 * This file is part of the Internet Configuration system and
 * is placed in the public domain for the benefit of all.
 *)

(* End Header *)

(* This file contains the core implementation of the Internet Config API.
	It defines the format of the ICRRecord type (which the outside world
	sees indirectly by way of the ICInstance, a pointer to this record) and
	the various IC API operations on that type.
*)

interface

	uses
		Types,
		Files, 
		QuickDraw, 
		Aliases, 
		Memory,
		Components,
		
		ICTypes,
		ICKeys;

	(* ***** Core Data Structure ***** *)

	(* The ICRRecord is the core data structure behind IC.  It holds all
		of the information needed to operate an instance, ie a connection
		to the IC system.
		
		The ICRRecord is instantiated in two different places.  Firstly,
		each component instance has a copy of this data structure.  The
		fields in the data structure are all used.  Most importantly,
		the instance field is used to hold the component's "target" instance.

		If the application talks directly to the component API (ie ICCAPI.p)
		then an ICInstance is a ComponentInstance and this is the
		only ICRRecord that comes into play.

		However, if an application uses the 68K glue and doesn't talk
		directly to the component, the 68K glue creates a copy of this
		record too.  This copy is used in one of two ways: if the component
		is installed, ICRRecord.instance contains the component instance
		for it and the other fields of the record are ignored.  If the
		component isn't installed, ICRRecord.instance is nil and the
		glue's link-in implementation comes into play.  The other fields
		in ICRRecord are then used to hold the link-in implementation's
		instance variables.
		
		The other part of the IC equation is the switch glue.  This is the
		glue that the application actually calls (in ICAPI.p) and this
		contains a bunch of routines that go:
		
			if ICRRecord.instance is nil
				call the link-in implementation
			else
				call the component
			end-if
		
		The fact that the component stores its target in its
		ICRRecord.instance field is a subtle but critical implementation
		detail.  Components, when they call themselves (ie a component
		routine that calls another component routine that is exported),
		are supposed to call through the current target.  This allows
		another component to set itself as the target and change or
		capture specific behaviour.
		
		For example, say I want to change the behaviour of ICParseURL.
		I could write an override component that reimplements
		ICParseURL and all client calls to ICParseURL would come to
		me for modification.  But when the client calls ICLaunchURL
		and the component's internal implementation calls ICParseURL,
		it won't necessarily come to me.  In order to make this work, the
		component is required to make these internal calls through the target.
		
		IC does this by having the link-in implementation call the IC API
		(ICAPI.p) when it needs to call itself.  Because a pointer to
		an ICRRecord is a valid ICInstance, it can just do this by
		taking the address of the inst parameter and casting it to
		an ICInstance.  Now there are two cases.  If the link-in
		implementation is being run stand-alone, ie it's hard-linked
		into an application and is not using the component, the
		ICRRecord.instance pointer will be nil and the switch glue will
		route the call back to the link-in implementation.  Alternatively,
		if the link-in implementation is being run as part of the
		component, the ICRRecord.instance will be the component's
		target, and the switch glue will call that target with the
		routines in ICCAPI.p.  Normally this will be the component
		itself, and we'll end up back in the component's link-in
		implementation.  But if the component has been overridden, we'll
		go to the override component.
		
		Cool huh?
		
		The ICRRecord is *completely* private to the implementation!  Do not rely
		on the existence or placement of these fields.
	*)
		
	type
		ICRRecord =
			record
				instance: ComponentInstance;
				haveConfigFile: Boolean;
				waitingForFirstBegin: Boolean;
				thisFileModified: Boolean;
				configFileSpec: FSSpec;
				configFileRefNum: integer;
				configFilePerm: ICPerm;
				defaultFileName: Str63;
				promptString: Str255;
			end;
		ICRRecordPtr = ^ICRRecord;
	
	(* The fields of the ICRRecord include:
	
		instance -- See the description above.
		haveConfigFile -- A Boolean that says whether this instance has been
			configured, ie we have decided on a definitive config file.  Specifically
			this means that configFileSpec is valid.
		waitingForFirstBegin -- A Boolean that says whether someone has
			called Begin on this config.  This is set every time we change
			config files, and cleared every time Begin is called.  The component
			uses this Boolean to tell whether this is the first Begin, in which
			cases it sanity checks the config file before calling through
			to the basic implementation.
		thisFileModified -- A Boolean that says whether this config file has
			been modified.  This is set every time SetPref is called, and
			cleared every time a new config file is specified.  The component
			uses this to determine whether it should save a new backup of
			this configuration file.
		configFileSpec -- A file spec that points to the current configuration
			file.  Only valid if haveConfigFile is set.
		configFileRefNum -- A Resource Manager refNum for the current
			config file.  This is 0 if the file isn't open.  When someone
			calls Begin, the file is opened and this is set to the file refNum.
			When they call End, the opposite happens.  The one exception
			is that the refNum can be 0 if someone calls Begin asking for read-only
			permissions on a file that doesn't exist yet.  For this reason,
			you use the next field to tell you whether you're inside a
			a Begin/End pair.
		configFilePerm -- The access permissions on the current file.  If it's not
			icNoPerm, you're inside a Begin/End pair.
		defaultFileName -- The default file name for the Internet
			Preferences file.  This is stored in the instance for historical
			reasons.  In the current environment, this code should simply
			call GetDefaultFileName to get it.
		promptString -- This is stored because there is not IC accessor
			routine that gets this value.  So when the link-in implementation
			needs to know this, it doesn't know whether to get it
			out of the resource fork (if it's running in the component)
			or use a hard-wired string (if it's running stand-alone).  I will
			most probably provide an accessor sometime soon.
	*)

	(* File Format Notes
		----------------
		The Internet Preferences file format is very simple.  IC is basically
		a thin layer over the Resource Manager.  All IC preferences are
		stored as resources of type kICPrefResourceType.  The key for
		the preference is stored as the resource.  The attributes for
		the preference are stored as the first 4 bytes of the resource.
		The resource ID is not significant.
		
		In IC 1.4 or later, the file also profits a version number, which
		is the version number of the component that created the default
		preferences in the file.  This version number is not set up by
		the link-in glue.  If a future component successfully updates the
		default preferences, this version will be set to the component's
		version.
	*)

	const
		kICPrefResourceType = 'ICRP';

	(* Copied Key Notes

		With version 1.4 and up, the ICAPI no longer requires that key
		strings be on word boundaries on 68000 machines.  This was 
		causing a problem for CodeWarrior C uses, because their
		compiler didn't align Pascal strings on word boundaries (duh!).
		Anyway, one of the tasks for 1.4 was to reduce the stack space
		usage, and fixing this key alignment problem just fell out
		of the wash.
		
		The external IC routines all use a Str255 interface, but
		the internal routines all use StringPtr instead.  This
		avoids the data copying.  When I need to use the string,
		I copy it into a local Str255 (typically called "copiedKey")
		using BlockMoveData.  BlockMoveData has no alignment 
		restrictions, so our 68000 machines are happy.
		
		In the process of doing this, I also guard myself against
		the key StringPtr being nil.  If any of my C programming
		friends try to do that, they'll get a nice error code.
	*)

	(* The remainer of these file is automagically generated by the
		"IC Interface Generator" HyperCard stack.  See the ICAPI.p
		file for a real description of what these API routines do.
	*)
	
(* Start ICRAPI.p *)

  
  (* ***** Starting Up and Shutting Down ***** *)
  
  function ICRStart(var inst : ICRRecord; creator : OSType) : ICError;
  function ICRStop(var inst : ICRRecord) : ICError;
  
  (* ***** Specifying a Configuration ***** *)
  
  function ICRFindConfigFile(var inst : ICRRecord; count : integer; folders : ICDirSpecArrayPtr) : ICError;
  function ICRFindUserConfigFile(var inst : ICRRecord; var where : ICDirSpec) : ICError;
  function ICRGeneralFindConfigFile(var inst : ICRRecord; searchPrefs : Boolean; canCreate : Boolean; count : integer; folders : ICDirSpecArrayPtr) : ICError;
  function ICRChooseConfig(var inst : ICRRecord) : ICError;
  function ICRChooseNewConfig(var inst : ICRRecord) : ICError;
  function ICRGetConfigName(var inst : ICRRecord; longname : Boolean; var name : Str255) : ICError;
  function ICRGetConfigReference(var inst : ICRRecord; ref : ICConfigRefHandle) : ICError;
  function ICRSetConfigReference(var inst : ICRRecord; ref : ICConfigRefHandle; flags : longint) : ICError;
  function ICRSpecifyConfigFile(var inst : ICRRecord; var config : FSSpec) : ICError;
  
  (* ***** Getting Information ***** *)
  
  function ICRGetSeed(var inst : ICRRecord; var seed : longint) : ICError;
  function ICRGetPerm(var inst : ICRRecord; var perm : ICPerm) : ICError;
  function ICRDefaultFileName(var inst : ICRRecord; var name : Str63) : ICError;
  
  (* ***** Reading and Writing Preferences ***** *)
  
  function ICRBegin(var inst : ICRRecord; perm : ICPerm) : ICError;
  function ICRGetPref(var inst : ICRRecord; key : StringPtr; var attr : ICAttr; buf : Ptr; var size : longint) : ICError;
  function ICRSetPref(var inst : ICRRecord; key : StringPtr; attr : ICAttr; buf : Ptr; size : longint) : ICError;
  function ICRFindPrefHandle(var inst : ICRRecord; key : StringPtr; var attr : ICAttr; prefh : Handle) : ICError;
  function ICRGetPrefHandle(var inst : ICRRecord; key : StringPtr; var attr : ICAttr; var prefh : Handle) : ICError;
  function ICRSetPrefHandle(var inst : ICRRecord; key : StringPtr; attr : ICAttr; prefh : Handle) : ICError;
  function ICRCountPref(var inst : ICRRecord; var count : longint) : ICError;
  function ICRGetIndPref(var inst : ICRRecord; n : longint; var key : Str255) : ICError;
  function ICRDeletePref(var inst : ICRRecord; key : StringPtr) : ICError;
  function ICREnd(var inst : ICRRecord) : ICError;
  
  (* ***** User Interface Stuff ***** *)
  
  function ICREditPreferences(var inst : ICRRecord; key : StringPtr) : ICError;
  
  (* ***** URL Handling ***** *)
  
  function ICRParseURL(var inst : ICRRecord; hint : StringPtr; data : Ptr; len : longint; var selStart : longint; var selEnd : longint; url : Handle) : ICError;
  function ICRLaunchURL(var inst : ICRRecord; hint : StringPtr; data : Ptr; len : longint; var selStart : longint; var selEnd : longint) : ICError;
  
  (* ***** Mappings Routines *****
   * 
   * Routines for interrogating mappings database.
   * 
   * ----- High Level Routines -----
   *)
  
  function ICRMapFilename(var inst : ICRRecord; filename : StringPtr; var entry : ICMapEntry) : ICError;
  function ICRMapTypeCreator(var inst : ICRRecord; fType : OSType; fCreator : OSType; filename : StringPtr; var entry : ICMapEntry) : ICError;
  
  (* ----- Mid Level Routines ----- *)
  
  function ICRMapEntriesFilename(var inst : ICRRecord; entries : Handle; filename : StringPtr; var entry : ICMapEntry) : ICError;
  function ICRMapEntriesTypeCreator(var inst : ICRRecord; entries : Handle; fType : OSType; fCreator : OSType; filename : StringPtr; var entry : ICMapEntry) : ICError;
  
  (* ----- Low Level Routines ----- *)
  
  function ICRCountMapEntries(var inst : ICRRecord; entries : Handle; var count : longint) : ICError;
  function ICRGetIndMapEntry(var inst : ICRRecord; entries : Handle; ndx : longint; var pos : longint; var entry : ICMapEntry) : ICError;
  function ICRGetMapEntry(var inst : ICRRecord; entries : Handle; pos : longint; var entry : ICMapEntry) : ICError;
  function ICRSetMapEntry(var inst : ICRRecord; entries : Handle; pos : longint; var entry : ICMapEntry) : ICError;
  function ICRDeleteMapEntry(var inst : ICRRecord; entries : Handle; pos : longint) : ICError;
  function ICRAddMapEntry(var inst : ICRRecord; entries : Handle; var entry : ICMapEntry) : ICError;

(* End ICRAPI.p *)

implementation

	uses
		Resources,
		GestaltEqu, 
		OSUtils, 
		ToolUtils, 
		StandardFile, 
		AppleTalk, 
		Folders, 
		Processes, 
		Errors, 
		AppleEvents, 
		LowMem,

		ICAPI, 
		ICLinkInSubs;

	const
		kICOurManufacturer = 'JPQE';

	function CloseIfOpen (var inst: ICRRecord): Boolean;
		forward;

	(* ***** Starting Up and Shutting Down ***** *)

	function ICRStart (var inst: ICRRecord; creator: OSType): ICError;
		(* See comment in ICAPI.p. *)
		var
			junk: ICError;
	begin
		{$unused creator}
		
		(* Initialise all the fields in the instance.  Some fields are
			initialised through resources.  We call API routines to do
			this because we don't know whether we're running as
			part of the component (and have resources) or as
			standalone glue.

			Note: In the component version, the promptString will be overwritten 
			by a resource, guaranteeing internationalisation.
		*)
		inst.haveConfigFile := false;
		inst.waitingForFirstBegin := false;
		inst.thisFileModified := false;
		inst.configFileSpec.vRefNum := 0;
		inst.configFileSpec.parID := 0;
		inst.configFileSpec.name := '';
		inst.configFileRefNum := 0;
		inst.configFilePerm := icNoPerm;
		junk := ICDefaultFileName(ICInstance(@inst), inst.defaultFileName);
		inst.promptString := 'Create configuration as:';
		
		ICRStart := noErr;
	end; (* ICRStart *)

	function ICRStop (var inst: ICRRecord): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: OSErr;
	begin
		if CloseIfOpen(inst) then begin
			err := paramErr;
		end else begin
			err := noErr;
		end; (* if *)
		ICRStop := err;
	end; (* ICRStop *)

	(* ***** Specifying a Configuration ***** *)

	function ValidDirSpec (folder: ICDirSpec): ICError;
		(* Returns noErr if the ICDirSpec passed in points to an existing directory.
			This routine is called to check that the ICDirSpecs passed into the various
			configuration routines are valid before we accept them.
		*)
		var
			cpb: CInfoPBRec;
	begin
		cpb.ioVRefNum := folder.vRefNum;
		cpb.ioDirID := folder.dirID;
		cpb.ioNamePtr := nil;
		cpb.ioFDirIndex := -1;
		ValidDirSpec := PBGetCatInfoSync(@cpb);
	end; (* ValidDirSpec *)

	function ScanFolder (var inst: ICRRecord; folder: ICDirSpec; var fileWeFound: FSSpec): Boolean;
		(* Scans a folder specified by the ICDirSpec looking for a valid Internet Preferences
			file.  The algorithm is a) first search for a file called Internet Preferences,
			b) then scan the entire folder looking for a type/creator match, and c)
			each time you check a file, you should resolve it if it's an alias.
		*)

		function FoundFile (folder: ICDirSpec; fileIndex: integer; var fileWeFound: FSSpec): OSErr;
			(* Check the file pointed to by either:

				a) if fileIndex = 0, folder.vRefNum + folder.dirID + fileWeFound.name
				b) if fileIndex > 0, folder.vRefNum + folder.dirID + fileIndex

				If the routine succeeds it returns noErr and fileWeFound is set to
				the file we found.  If the routine fails, it either returns the real
				error code, or a synthetic error code of 1, which means keep 
				incrementing the index and call me again.
			*)
			var
				err: OSErr;
				cpb: CInfoPBRec;
				isFolder: Boolean;
				junkWasAlias: Boolean;
		begin
			cpb.ioVRefNum := folder.vRefNum;
			cpb.ioDirID := folder.dirID;
			cpb.ioNamePtr := @fileWeFound.name;
			cpb.ioFDirIndex := fileIndex;
			err := PBGetCatInfoSync(@cpb);
			if err = noErr then begin
				
				(* If we get no error, we have to decide whether this is a useful file
					or not.  If it is, we continue to return noErr and that's the file we
					return.  If it isn't, we return 1, which tells the client to keep
					incrementing the index.
				*)
				fileWeFound.vRefNum := cpb.ioVRefNum;
				fileWeFound.parID := cpb.ioFlParID;
				if (btst(cpb.ioFlAttrib, 4) or (cpb.ioFlFndrInfo.fdType <> ICfiletype)) then begin
					err := 1; 				(* Folder, or not of the right type. *)
				end else if ICUHaveAliasManager then begin
					err := ResolveAliasFile(fileWeFound, true, isFolder, junkWasAlias);
					if (err <> noErr) or isFolder then begin
						err := 1;			(* Broken alias, or an alias to a folder. *)
					end; (* if *)
				end; (* if *)
			end; (* if *)
			FoundFile := err;
		end; (* FoundFile *)

		var
			err: ICError;
			found: Boolean;
			fileIndex: integer;
	begin
		(* First look for the Internet Preferences file in the folder. *)
		fileWeFound.name := inst.defaultFileName;
		found := (FoundFile(folder, 0, fileWeFound) = noErr);
		
		(* If we didn't find it, scan through all the files in the folder looking
			for a matching file type.
		*)
		if not found then begin
			fileIndex := 1;
			repeat
				fileWeFound.name := '';
				err := FoundFile(folder, fileIndex, fileWeFound);
				fileIndex := fileIndex + 1;
			until err <> 1;
			found := (err = noErr);
		end; (* if *)
		ScanFolder := found;
	end; (* ScanFolder *)

	(* ***** Specifying a Configuration ***** *)

	function ICRFindConfigFile (var inst: ICRRecord; count: integer; folders: ICDirSpecArrayPtr): ICError;
		(* See comment in ICAPI.p. *)
	begin
		ICRFindConfigFile := ICGeneralFindConfigFile(ICInstance(@inst), 
																				true, true, 
																				count, folders);
	end; (* ICRFindConfigFile *)

	function ICRFindUserConfigFile (var inst: ICRRecord; var where: ICDirSpec): ICError;
		(* See comment in ICAPI.p. *)
	begin
		ICRFindUserConfigFile := ICGeneralFindConfigFile(ICInstance(@inst), 
																				false, true, 
																				1, @where);
	end; (* ICRFindUserConfigFile *)

	function ICRGeneralFindConfigFile (var inst: ICRRecord; searchPrefs: Boolean; canCreate: Boolean; count: integer; folders: ICDirSpecArrayPtr): ICError;
		(* See comment in ICAPI.p. *)

		function FindPrefFolder (var prefFolder: ICDirSpec): OSErr;
			(* Find the preferences folder in a System 6 + 7 friendly fashion.
				If we have FindFolder, call it directly.  [ICUFindFolder is our
				own 68K inline for FindFolder, guaranteed to avoid any
				glue provided by the development environment.]
				If we don't, call SysEnvirons to find the System Folder
				and use it.
			*)
			var
				err: OSErr;
				env: SysEnvRec;
				junk: longint;
		begin
			if ICUHaveFindFolder then begin
				(* Gestalt says it's implemented -- call it directly *)
				err := ICUFindFolder(kOnSystemDisk, kPreferencesFolderType, true, 
													prefFolder.vRefNum, prefFolder.dirID);
			end else begin
				(* Simulate FindFolder by calling SysEnvirons to get the System Folder
					and returning it.
				*)
				err := SysEnvirons(curSysEnvVers, env);
				if err = noErr then begin
					err := GetWDInfo(env.sysVRefNum, prefFolder.vRefNum, prefFolder.dirID, junk);
				end; (* if *)
			end; (* if *)
			FindPrefFolder := err;
		end; (* FindPrefFolder *)

		var
			err: ICError;
			dirSpecIndex: integer;
			found: Boolean;
			prefFolder: ICDirSpec;
			lastFolderScanned: ICDirSpec;
			tempConfigFileSpec: FSSpec;
	begin
		(* Check the parameters. *)
		err := noErr;
		if (err = noErr) & (inst.configFilePerm <> icNoPerm) then begin
			err := paramErr;					(* Must not be inside Begin/End pair. *)
		end; (* if *)
		if (err = noErr) & ((count < 0) | ((count <> 0) & (folders = nil))) then begin
			err := paramErr;					(* Check that count and folders makes sense. *)
		end; (* if *)
		if (err = noErr) & (count = 0) & not searchPrefs & canCreate then begin
			err := paramErr;					(* Being told to create a file without specifying where. *)
		end; (* if *)
		(* Check that all the specified directories exist. *)
		dirSpecIndex := 0;
		while (err = noErr) & (dirSpecIndex < count) do begin
			err := ValidDirSpec(folders^[dirSpecIndex]);
			dirSpecIndex := dirSpecIndex + 1;
		end; (* for *)
		
		(* OK, parameters are cool, let's start the job. *)
		if err = noErr then begin
		
			(* Search all the specified directories. *)
			dirSpecIndex := 0;
			found := false;
			while (dirSpecIndex < count) and not found do begin
				found := ScanFolder(inst, folders^[dirSpecIndex], tempConfigFileSpec);
				lastFolderScanned := folders^[dirSpecIndex];
				dirSpecIndex := dirSpecIndex + 1;
			end; (* while *)
			
			(* Now, if we haven't found it yet and we're asked to,
				search the preferences folder.
			*)
			if not found & searchPrefs then begin
				err := FindPrefFolder(prefFolder);
				if (err = noErr) then begin
					found := ScanFolder(inst, prefFolder, tempConfigFileSpec);
					lastFolderScanned := prefFolder;
				end;
			end; (* if *)
			
			(* Finally, if we still can't find it and we're allowed to, create the
				file in the last folder we scanned.
			*)
			if not found and canCreate then begin
				tempConfigFileSpec.vRefNum := lastFolderScanned.vRefNum;
				tempConfigFileSpec.parID := lastFolderScanned.dirID;
				tempConfigFileSpec.name := inst.defaultFileName;
				found := true;
			end; (* if *)
			
			(* If we still can't find it, return an error. *)
			if not found then begin
				err := icConfigNotFoundErr;
			end; (* if *)
		end; (* if *)
		
		(* Now pass what we found along to ICSpecifyConfigFile. *)
		if err = noErr then begin
			err := ICSpecifyConfigFile(ICInstance(@inst), tempConfigFileSpec);
		end; (* if *)
		inst.haveConfigFile := (err = noErr);
		ICRGeneralFindConfigFile := err;
	end; (* ICRGeneralFindConfigFile *)

	(* ***** Standard File Utilities ***** *)

	procedure SetSFCWD (var inst: ICRRecord);
		(* Point Standard File at the directory that contains the instance's current
			config file, if there is one.
			
			Note that this most probbaly fails with newer (post 7.5) versions of
			Standard File, but I'm not sure whether it's worth doing the workaround
			at this time.
		*)
	begin
		if inst.haveConfigFile then begin
			LMSetCurDirStore( inst.configFileSpec.parID );
			LMSetSFSaveDisk( -inst.configFileSpec.vRefNum );
		end; (* if *)
	end; (* SetSFCWD *)

	function GetFile (var inst: ICRRecord; var chosenFileSpec: FSSpec): ICError;
		(* Prompt the user to choose a file using the System 6 or 7
			SF dialogs.
		*)
		var
			err: ICError;
			typeList: SFTypeList;
			newReply: StandardFileReply;
			oldReply: SFReply;
			junkLong: longInt;
	begin
		SetSFCWD(inst);
		err := userCanceledErr;
		typeList[0] := ICfiletype;
		if ICUHaveNewStandardFile then begin
			StandardGetFile(nil, 1, @typeList, newReply);
			if newReply.sfGood then begin
				chosenFileSpec := newReply.sfFile;
				err := noErr;
			end; (* if *)
		end else begin
			SFGetFile(Point($4040), '', nil, 1, @typeList, nil, oldReply);
			if oldReply.good then begin
				err := GetWDInfo(oldReply.vRefNum, chosenFileSpec.vRefNum, chosenFileSpec.parID, junkLong);
				chosenFileSpec.name := oldReply.fName;
			end; (* if *)
		end; (* if *)
		GetFile := err;
	end; (* GetFile *)

	function PutFile (var inst: ICRRecord; var chosenFileSpec: FSSpec): ICError;
		(* Prompt the user to create a new file using the System 6 or 7
			SF dialogs.
		*)
		var
			err: ICError;
			typeList: SFTypeList;
			newReply: StandardFileReply;
			oldReply: SFReply;
			junkLong: longInt;
			defaultFileName: Str63;
	begin
		SetSFCWD(inst);
		err := userCanceledErr;
		typeList[0] := ICfiletype;
		
		(* Setup the prompt, either with the name of the current file or the default
			file name.
		*)
		if inst.haveConfigFile then begin
			defaultFileName := inst.configFileSpec.name;
		end else begin
			defaultFileName := inst.defaultFileName;
		end; (* if *)
		
		if ICUHaveNewStandardFile then begin
			StandardPutFile(inst.promptString, defaultFileName, newReply);
			if newReply.sfGood then begin
				chosenFileSpec := newReply.sfFile;
				err := noErr;
			end; (* if *)
		end else begin
			SFPutFile(Point($4040), inst.promptString, defaultFileName, nil, oldReply);
			if oldReply.good then begin
				err := GetWDInfo(oldReply.vRefNum, chosenFileSpec.vRefNum, chosenFileSpec.parID, junkLong);
				chosenFileSpec.name := oldReply.fName;
			end; (* if *)
		end; (* if *)
		PutFile := err;
	end; (* PutFile *)

	(* ***** Choosing a Configuration ***** *)

	function ICRChooseConfig (var inst: ICRRecord): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: OSErr;
			chosenFileSpec: FSSpec;
	begin
		(* Check parameters. *)
		err := noErr;
		if (err = noErr) & (inst.configFilePerm <> icNoPerm) then begin
			err := paramErr;					(* Must not be inside Begin/End pair. *)
		end; (* if *)
		
		(* Check whether we can pose dialogs, ask the user to specify a file,
			then specify that config file.
		*)
		if err = noErr then begin
			err := ICUCanInteract;
		end; (* if *)
		if err = noErr then begin
			err := GetFile(inst, chosenFileSpec);
		end; (* if *)
		if err = noErr then begin
			err := ICSpecifyConfigFile(ICInstance(@inst), chosenFileSpec);
		end; (* if *)
		ICRChooseConfig := err;
	end; (* ICRChooseConfig *)

	function ICRChooseNewConfig (var inst: ICRRecord): ICError;
		(* See comment in ICAPI.p. *)
		var
			err, junk: OSErr;
			chosenFileSpec: FSSpec;
	begin
		(* Check parameters. *)
		err := noErr;
		if (err = noErr) & (inst.configFilePerm <> icNoPerm) then begin
			err := paramErr;					(* Must not be inside Begin/End pair. *)
		end;

		(* Check whether we can pose dialogs, ask the user to specify a new file,
			delete and recreate the file, then specify that config file.
		*)
		if err = noErr then begin
			err := ICUCanInteract;
		end; (* if *)
		if err = noErr then begin
			err := PutFile(inst, chosenFileSpec);
		end; (* if *)
		if err = noErr then begin
			junk := HDelete(chosenFileSpec.vRefNum, chosenFileSpec.parID, chosenFileSpec.name);
			err := HCreate(chosenFileSpec.vRefNum, chosenFileSpec.parID, chosenFileSpec.name, ICcreator, ICfiletype);
		end; (* if *)
		if err = noErr then begin
			err := ICSpecifyConfigFile(ICInstance(@inst), chosenFileSpec);
		end; (* if *)
		ICRChooseNewConfig := err;
	end; (* ICRChooseNewConfig *)

	(* ***** Other Configuration API Routines ***** *)

	function ICRGetConfigName (var inst: ICRRecord; longname: Boolean; var name: Str255): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: OSErr;
	begin
		(* Check parameters. *)
		err := noErr;
		if not inst.haveConfigFile then begin
			err := paramErr;					(* Must have a configuration. *)
		end; (* if *)

		if longname then begin
			err := ICUFSSPecToFullPath(inst.configFileSpec, name);
		end else begin
			name := inst.configFileSpec.name;
			err := noErr;
		end; (* if *)
		ICRGetConfigName := err;
	end; (* ICRGetConfigName *)

	function ICRGetConfigReference (var inst: ICRRecord; ref: ICConfigRefHandle): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			configRefHeader: ICConfigRef;
			junkLong: longint;
	begin
		err := noErr;
		if not inst.haveConfigFile then begin
			err := paramErr;					(* Must have a configuration. *)
		end; (* if *)
		if (err = noErr) & (ref = nil) then begin
			err := paramErr;					(* Client must supply handle for results. *)
		end; (* if *)
		
		(* First create an ICFileSpec to the file.  This contains both a real alias
			and a poor man's alias for System 6 operation.
		*)
		if err = noErr then begin
			err := FSSpecToICFileSpec(inst.configFileSpec, ICFileSpecHandle(ref));
		end; (* if *)
		
		(* Now insert our manufacturer code at the front of the ICFileSpec
			to make it a config reference.
		*)
		if err = noErr then begin
			configRefHeader.manufacturer := kICOurManufacturer;
			junkLong := Munger(Handle(ref), 0, nil, 0, @configRefHeader, sizeof(configRefHeader));
			err := MemError;
		end; (* if *)
		
		(* Clean up. *)
		if err <> noErr then begin
			SetHandleSize(Handle(ref), 0);
		end; (* if *)
		ICRGetConfigReference := err;
	end; (* ICRGetConfigReference *)

	function ICRSetConfigReference (var inst: ICRRecord; ref: ICConfigRefHandle; flags: longint): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			filespec: ICFileSpecHandle;
			junkLong: longInt;
			newConfigFileSpec: FSSpec;
	begin
		(* Check parameters. *)
		err := noErr;
		if inst.configFilePerm <> icNoPerm then begin
			err := paramErr;					(* Must not be inside Begin/End pair. *)
		end; (* if *)
		if (err = noErr) & (ref = nil) then begin
			err := paramErr;					(* Client must supply handle for results. *)
		end; (* if *)
		if (err = noErr) & (GetHandleSize(Handle(ref)) < 4) then begin
			err := paramErr;					(* Handle must be a reasonable size. *)
		end; (* if *)
		if (err = noErr) & (ref^^.manufacturer <> kICOurManufacturer) then begin
			err := icConfigInappropriateErr;	(* Config ref header must indicate that the config ref is ours. *)
		end; (* if *)
		if (err = noErr) & (GetHandleSize(Handle(ref)) < sizeof(ICConfigRef) + sizeof(ICFileSpec)) then begin
			err := paramErr;					(* Config ref data must be big enough for minimal ICFileSpec *)
		end; (* if *)

		(* Now do the work. *)
		if err = noErr then begin
			(* Make a copy of the handle. *)
			filespec := ICFileSpecHandle(ref);
			err := HandToHand(Handle(filespec));
			if err = noErr then begin
			
				(* Delete the config ref header off the front of the handle. *)
				junkLong := Munger(Handle(filespec), 0, nil, sizeof(ICConfigRef), @junkLong, 0);

				(* Pass the remaining handle to our 'resolve alias' code. *)
				err := ICFileSpecToFSSpec(filespec, not btst(flags, icNoUserInteraction_bit), newConfigFileSpec);
				if err = fnfErr then begin
					(* Swallow fnfErr because it might be a reference to a file that
						had never been created in the first place.
					*)
					err := noErr;
				end; (* if *)
				DisposeHandle(Handle(filespec));
			end; (* if *)
		end; (* if *)

		(* Use the new config file. *)
		if err = noErr then begin
			err := ICSpecifyConfigFile(ICInstance(@inst), newConfigFileSpec);
		end; (* if *)
		ICRSetConfigReference := err;
	end; (* ICRSetConfigReference *)

	function ICRSpecifyConfigFile (var inst: ICRRecord; var config: FSSpec): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			folder: ICDirSpec;
	begin
		(* Check parameters. *)
		err := noErr;
		if inst.configFilePerm <> icNoPerm then begin
			err := paramErr;					(* Must not be inside Begin/End pair. *)
		end; (* if *)
		
		(* Do the job.  Check that the new file is potentially valid (ie the file
			may not exist, but the directory it's in must and then switch
			configurations.
		*)
		if err = noErr then begin
			folder.vRefNum := config.vRefNum;
			folder.dirID := config.parID;
			err := ValidDirSpec(folder);
			if err = noErr then begin
				inst.configFileSpec := config;
				inst.waitingForFirstBegin := true;
				inst.thisFileModified := false;
			end; (* if *)
			inst.haveConfigFile := (err = noErr);
		end; (* if *)
		ICRSpecifyConfigFile := err;
	end; (* ICRSpecifyConfigFile *)

	(* ***** Getting Information ***** *)

	function ICRGetSeed (var inst: ICRRecord; var seed: longint): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			cpb: CInfoPBRec;
	begin
		seed := 0;
		err := fnfErr;
		if inst.haveConfigFile then begin
			cpb.ioVRefNum := inst.configFileSpec.vRefNum;
			cpb.ioDirID := inst.configFileSpec.parID;
			cpb.ioNamePtr := @inst.configFileSpec.name;
			cpb.ioFDirIndex := 0;
			err := PBGetCatInfoSync(@cpb);
			if err = noErr then begin
				seed := cpb.ioFlMdDat;
			end else if err = fnfErr then begin
				err := noErr;
			end; (* if *)
		end; (* if *)
		ICRGetSeed := err;
	end; (* ICRGetSeed *)

	function ICRGetPerm (var inst: ICRRecord; var perm: ICPerm): ICError;
		(* See comment in ICAPI.p. *)
	begin
		perm := inst.configFilePerm;
		ICRGetPerm := noErr;
	end; (* ICRGetPerm *)

	function ICRDefaultFileName (var inst: ICRRecord; var name: Str63): ICError;
		(* See comment in ICAPI.p.  This routine is typically overridden by
			the component to get the value from a resource and thus be localisable.
		*)
	begin
		{$unused inst}
		name := ICdefault_file_name;
		ICRDefaultFileName := noErr;
	end; (* ICRDefaultFileName *)

  (* ***** Begin/End Utilities ***** *)

	function PermToFSPerm (perm: ICPerm): integer;
		(* Maps an ICPerm to an file system permission. *)
	begin
		case perm of
			icReadOnlyPerm: 
				PermToFSPerm := fsRdPerm;
			icReadWritePerm: 
				PermToFSPerm := fsRdWrPerm;
			otherwise
				PermToFSPerm := 0;
		end; (* case *)
	end; (* PermToFSPerm *)

	function CheckInside (var inst: ICRRecord): ICError;
		(* The routine returns an error code based on whether you're
			inside a Begin/End pair. CheckInside is a utility used by API
			routines that require you to be inside a Begin/End pair to operate.
		*)
	begin
		if inst.configFilePerm = icNoPerm then begin
			CheckInside := paramErr;
		end else begin
			CheckInside := noErr;
		end; (* if *)
	end; (* CheckInside *)

	function ForceInside (var inst: ICRRecord; perm: ICPerm; var forceInsideInfo: Boolean): ICError;
		(* This routine makes sure we're currently inside a Begin/End pair,
			calling Begin if necessary.  The routine sets forceInsideInfo to
			true if it called Begin, so that ReleaseInside knows whether
			to call End.  ForceInside is a utility used by API routines
			that automagically call Begin for you, such as GetPref.
		*)
		var
			err: ICError;
	begin
		forceInsideInfo := false;
		if (inst.configFilePerm = perm) or			(* We already have the required permission OR *)
				 ((inst.configFilePerm = icReadWritePerm) 		(* We asking for read AND *)
				 			and (perm = icReadOnlyPerm))							(* We have read/write. *)
				 					then begin
			(* The permissions already match up, do nothing. *)
			err := noErr;
		end else if inst.configFilePerm = icNoPerm then begin
			(* We have no existing permissions, call Begin and record whether
				we need to call End.
			*)
			err := ICBegin(ICInstance(@inst), perm);
			forceInsideInfo := (err = noErr);
		end else begin
			(* If neither of the above cases work, fail with an error.  The only
				case that springs to mind is that we have read-only permissions
				and someone is asking for read/write.
			*)
			err := icPermErr;
		end; (* if *)
		ForceInside := err;
	end; (* ForceInside *)

	function ReleaseInside (var inst: ICRRecord; forceInsideInfo: Boolean): ICError;
		(* Undoes the effect of a ForceInside. Note that unlike Begin/End, you
			should can safely call this regardless of whether ForceInside returns
			an error.
		*)
	begin
		if forceInsideInfo then begin
			ReleaseInside := ICEnd(ICInstance(@inst));
		end else begin
			ReleaseInside := noErr;
		end; (* if *)
	end; (* ReleaseInside *)

	function CloseIfOpen (var inst: ICRRecord): Boolean;
		(* Close the instance's config file if it's open.  Returns
			true if the instance was in the right state for
			this routine to be called.
		*)
	begin
		CloseIfOpen := (inst.configFilePerm <> icNoPerm);
		if inst.configFileRefNum <> 0 then begin
			CloseResFile(inst.configFileRefNum);
			inst.configFileRefNum := 0;
		end; (* if *)
		inst.configFilePerm := icNoPerm;
	end; (* CloseIfOpen *)

	(* ***** Begin/End ***** *)

	function ICRBegin (var inst: ICRRecord; perm: ICPerm): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			resourceRefNum: integer;
			junk: OSErr;
	begin
		(* Check parameters. *)
		err := noErr;
		if (inst.configFilePerm <> icNoPerm) or (perm = icNoPerm) then begin
			err := paramErr;		(* Can't call Begin twice, or call it to ask for no permissions. *)
		end; (* if *)
		if err = noErr then begin
			if not inst.haveConfigFile then begin
				err := bdNamErr;	(* Need to know where the file is do anything. *)
			end; (* if *)
		end; (* if *)

		(* Open the resource file. *)
		if err = noErr then begin
			resourceRefNum := HOpenResFile(inst.configFileSpec.vRefNum, 
																inst.configFileSpec.parID, 
																inst.configFileSpec.name, 
																PermToFSPerm(perm));
			err := ResError;

			(* If the resource file does not exist (or its resource fork is empty), do
				some special processing.
			*)
			if (err = fnfErr) or (err = eofErr) then begin
				case perm of
					icReadOnlyPerm:
						begin
							(* Client asking for read-only, don't create the file, just
								mark the special case in the data that gets put in the instance.
							*)
							resourceRefNum := 0;
							err := noErr;
						end; (* icReadOnlyPerm *)
					icReadWritePerm:
						begin
							(* Client requested read/write, create and open the resoure file. *)
							junk := HCreate(inst.configFileSpec.vRefNum, 
														inst.configFileSpec.parID, 
														inst.configFileSpec.name, 
														ICcreator, 
														ICfiletype);
							HCreateResFile(inst.configFileSpec.vRefNum, 
														inst.configFileSpec.parID, 
														inst.configFileSpec.name);
							resourceRefNum := HOpenResFile(inst.configFileSpec.vRefNum, 
														inst.configFileSpec.parID, 
														inst.configFileSpec.name, 
														PermToFSPerm(perm));
							err := ResError;
						end; (* icReadWritePerm *)
				end; (* case *)
			end; (* if *)
		end; (* if *)

		(* All is cool, record the fact that we've opened the file in the instance. *)
		if err = noErr then begin
			inst.configFileRefNum := resourceRefNum;
			inst.configFilePerm := perm;
			inst.waitingForFirstBegin := false;
		end; (* if *)
		
		(* Clean up the error result. *)
		case err of
			opWrErr, permErr: 
				err := icNoMoreWritersErr;
			otherwise
				(* do nothing *)
		end; (* case *)
		ICRBegin := err;
	end; (* ICRBegin *)

	function ICREnd (var inst: ICRRecord): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			junkBool: Boolean;
	begin
		err := CheckInside(inst);
		junkBool := CloseIfOpen(inst);
		ICREnd := err;
	end; (* ICREnd *)

	(* ***** Basic Reading and Writing Preferences ***** *)

	function ICRGetPref (var inst: ICRRecord; key: StringPtr; var attr: ICAttr; buf: Ptr; var size: longint): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			err2: ICError;
			copiedKey: Str255;
			maxSizeOfClientBuffer: longint;
			trueSizeOfPrefResource: longint;
			currentRefNum: integer;
			prefResourceH: Handle;
			forceInsideInfo: Boolean;
	begin
		(* Initialise our locals and return results to nice default values. *)
		copiedKey := '';
		maxSizeOfClientBuffer := size;
		size := 0;
		attr := ICattr_no_change;
		prefResourceH := nil;

		(* Make sure we're inside a read capable Begin/End pair. *)
		err := ForceInside(inst, icReadOnlyPerm, forceInsideInfo);
		if (err = noErr) and (inst.configFileRefNum = 0) then begin
			(* If the config file is still not open, it means we've hit the special
				case of opening a missing file read-only.  There is no chance
				of such a file (there is no file!) containing preferences, so
				we can error the call now.
			*)
			err := icPrefNotFoundErr;
		end; (* if *)

		(* Check other parameters. *)
		if (err = noErr) and (key <> nil) then begin
			BlockMoveData(key, @copiedKey, ord(key^[0]) + 1);
		end; (* if *)
		if (err = noErr) and ((copiedKey = '') or ((maxSizeOfClientBuffer < 0) and (buf <> nil))) then begin
			(* key was nil, or key was empty, or maxSizeOfClientBuffer was bogus without buf being nil. *)
			err := paramErr;
		end; (* if *)

		(* Do the work. *)
		if err = noErr then begin

			(* Use our resource file.  We need to do this because the client might go:
				Begin, switch resource chain order, GetPref, so we're not guaranteed
				to be at the top of this point.
			*)
			currentRefNum := CurResFile;
			UseResFile(inst.configFileRefNum);
			err := ResError;
			if err = noErr then begin
				
				(* Get the existing preference and handle weird Resource Manager
					error conditions.
				*)
				prefResourceH := Get1NamedResource(kICPrefResourceType, copiedKey);
				err := ResError;
				if prefResourceH = nil then begin
					err := icPrefNotFoundErr;
				end; (* if *)
				
				(* Get and check the preference resource for validity. It must contain
					at least 4 bytes of attributes at the beginning.
				*)
				if err = noErr then begin
					trueSizeOfPrefResource := GetHandleSize(prefResourceH);
					if trueSizeOfPrefResource < sizeof(ICAttr) then begin
						err := icPrefDataErr;
					end; (* if *)
				end; (* if *)
				
				(* Return the preference to the client. *)
				if err = noErr then begin
					size := trueSizeOfPrefResource - sizeof(ICAttr);
					attr := longintPtr(prefResourceH^)^;
					if (buf <> nil) and (size <> 0) then begin
						if size > maxSizeOfClientBuffer then begin
							err := icTruncatedErr;
						end else begin
							maxSizeOfClientBuffer := size;
						end; (* if *)
						BlockMoveData(Ptr(longint(prefResourceH^) + sizeof(ICAttr)), buf, maxSizeOfClientBuffer);
					end; (* if *)
				end; (* if *)
				UseResFile(currentRefNum);
			end; (* if *)
		end; (* if *)
		
		(* Clean up. *)
		if prefResourceH <> nil then begin
			ReleaseResource(prefResourceH);
		end; (* if *)
		err2 := ReleaseInside(inst, forceInsideInfo);
		if err = noErr then begin
			err := err2;
		end; (* if *)
		ICRGetPref := err;
	end; (* ICRGetPref *)

	function ICRSetPref (var inst: ICRRecord; key: StringPtr; attr: ICAttr; buf: Ptr; size: longint): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			err2: ICError;
			copiedKey: Str255;
			oldAttr: longint;
			currentRefNum: integer;
			prefH: Handle;
			id: integer;
			forceInsideInfo: Boolean;
	begin
		(* Initialise our locals and return results to nice default values. *)
		copiedKey := '';
		prefH := nil;
		if buf = nil then begin
			(* API defines that if buf is nil, we're only setting the attributes
				and we should ignore size.  As we have a validity check for size
				later on, we want to make sure it'll pass that check.
			*)
			size := 0;
		end; (* if *)

		(* Make sure we're inside a read/write capable Begin/End pair. *)
		err := ForceInside(inst, icReadWritePerm, forceInsideInfo);
		if (err = noErr) and (inst.configFilePerm <> icReadWritePerm) then begin
			err := icPermErr;
		end; (* if *)
		if (err = noErr) and (inst.configFileRefNum = 0) then begin
			(* If the config file is still not open, it means we've hit the special
				case of opening a missing file read-only.  But we asked to open
				it read/write, so there's obviously something wrong.
			*)
			err := icInternalErr;
		end; (* if *)

		(* Check other parameters. *)
		if (err = noErr) and (key <> nil) then begin
			BlockMoveData(key, @copiedKey, ord(key^[0]) + 1);
		end; (* if *)
		if (err = noErr) and ((copiedKey = '') or (size < 0)) then begin
			(* key was nil, or key was empty, or size is bogus. *)
			err := paramErr;
		end; (* if *)
		
		(* Do the work. *)
		if err = noErr then begin

			(* Use our resource file.  We need to do this because the client might go:
				Begin, switch resource chain order, SetPref, so we're not guaranteed
				to be at the top of this point.
			*)
			currentRefNum := CurResFile;
			UseResFile(inst.configFileRefNum);
			err := ResError;
			if err = noErr then begin

				(* Make sure the preference exists and is vaguely valid, ie
					has a size greater than 4.  If it doesn't, trash that resource
					and pretend we never saw it.
				*)
				prefH := Get1NamedResource(kICPrefResourceType, copiedKey);
				if (prefH <> nil) & (GetHandleSize(prefH) < sizeof(ICAttr)) then begin { very bad! }
					RemoveResource(prefH);
					DisposeHandle(prefH);
					prefH := nil;
				end; (* if *)
				
				(* Work out the attributes that we're trying to set and check that
					this won't conflict with the permissions.
				*)
				if prefH = nil then begin
					oldAttr := 0;
				end else begin
					oldAttr := longintPtr(prefH^)^;
				end; (* if *)
				if attr = ICattr_no_change then begin
					attr := oldAttr;
				end; (* if *)
				if btst(oldAttr, ICattr_locked_bit) and
						 btst(attr, ICattr_locked_bit) and
						 		(buf <> nil) then begin
					err := icPermErr;
				end; (* if *)
				
				(* Note that we can't have an icPermErr if prefH is nil because
					prefH = nil => oldAttr = 0 => btst(oldAttr, ICattr_locked_bit) = false.
					So there's no possibility of the next chunk of code swallowing
					the error from the previous chunk of code.
				*)
				
				(* If we didn't find an old preference, create one and add it to
					the resource file, making sure we give it a unique ID.  Note
					that when we come out of this if statement, either
						o err = 0 and prefH is a valid resource handle, OR
						o err <> 0 and prefH is nil
				*)
				if prefH = nil then begin
					prefH := NewHandle(size + sizeof(ICAttr));
					err := MemError;
					if err = noErr then begin
						repeat
							id := Unique1ID(kICPrefResourceType);
						until id > 127;
						AddResource(prefH, kICPrefResourceType, id, copiedKey);
						err := ResError;
						if err <> noErr then begin
							DisposeHandle(prefH);
							prefH := nil;
						end; (* if *)
					end; (* if *)
				end; (* if *)

				(* Resize the preference handle to accomodate the new
					preference data.  Only do this if buf <> nil because the
					API says that buf=nil implies that we're only operating
					on the attributes not preferences data.
					Then copy the preference data into the new handle.
				*)
				if (err = noErr) & (buf <> nil) then begin
					SetHandleSize(prefH, size + sizeof(ICAttr));
					err := MemError;
				end; (* if *)
				if (err = noErr) & (size > 0) then begin
					BlockMoveData(buf, Ptr(longint(prefH^) + sizeof(ICAttr)), size);
				end; (* if *)
				
				(* If all goes well, copy in the new attributes and mark the resource
					as changed.
				*)
				if (err = noErr) then begin
					longintPtr(prefH^)^ := attr;
					ChangedResource(prefH);
					WriteResource(prefH);
					err := ResError;
				end; (* if *)
				UseResFile(currentRefNum);
			end; (* if *)
		end; (* if *)

		(* Clean up. *)
		if prefH <> nil then begin
			ReleaseResource(prefH);
		end; (* if *)
		err2 := ReleaseInside(inst, forceInsideInfo);
		if err = noErr then begin
			err := err2;
		end; (* if *)
		if err = noErr then begin
			inst.thisFileModified := true;
		end; (* if *)
		ICRSetPref := err;
	end; (* ICRSetPref *)

	(* ***** Reading and Writing Preference Handles ***** *)

	function ICRGetPrefHandle (var inst: ICRRecord; key: StringPtr; var attr: ICAttr; var prefh: Handle): ICError;
		(* See comment in ICAPI.p. This function is a deprecated form of FindPrefHandle.
			In this implementation, it provides a simple wrapper around FindPrefHandle
			that 'fixes' the behaviour.  Clients shouldn't really be calling this routine.
		*)
		var
			err: ICError;
	begin
		(* Allocates the handle in the current heap -- not a standard IC API practice. *)
		prefh := NewHandle(0);
		err := MemError;

		(* Call through to FindPrefHandle, which will do all the parameter validation. *)
		if err = noErr then begin
			err := ICFindPrefHandle(ICInstance(@inst), key^, attr, prefh);
		end; (* if *)
		
		(* If the preferences isn't found, return noErr and an empty handle. *)
		if err = icPrefNotFoundErr then begin
			SetHandleSize(prefh, 0);
			attr := 0;
			err := noErr;
		end; (* if *)
		ICRGetPrefHandle := err;
	end; (* ICRGetPrefHandle *)

	function ICRFindPrefHandle (var inst: ICRRecord; key: StringPtr; var attr: ICAttr; prefh: Handle): ICError;
		(* See comment in ICAPI.p.  This routine calls ForceInside to speed things up.
			ForceInside will do an Begin and hence open the resource file, which is
			good because otherwise GetPref would open it up each time we call it,
			ie twice.
		*)
		var
			err: ICError;
			err2: ICError;
			preferenceSize: longint;
			forceInsideInfo: Boolean;
			s : SignedByte;
	begin
		(* Check the parameters. *)
		err := noErr;
		if prefh = nil then begin
			err := paramErr;
		end; (* if *)

		(* Do the work. *)
		preferenceSize := 0;
		if err = noErr then begin
			err := ForceInside(inst, icReadOnlyPerm, forceInsideInfo);
			s := HGetState(prefh);
			
			(* Get the size of the preference, resize the handle to that size,
				and then get the preference data into the handle.
			*)
			if err = noErr then begin
				err := ICGetPref(ICInstance(@inst), key^, attr, nil, preferenceSize);
			end; (* if *)
			if err = noErr then begin
				HUnlock(prefh);
				SetHandleSize(prefh, preferenceSize);
				err := MemError;
			end; (* if *)
			if err = noErr then begin
				HLock(prefh);
				err := ICGetPref(ICInstance(@inst), key^, attr, prefh^, preferenceSize);
			end; (* if *)

			HSetState(prefh, s);
			err2 := ReleaseInside(inst, forceInsideInfo);
			if err = noErr then begin
				err := err2;
			end; (* if *)
		end; (* if *)

		(* Clean up. *)
		if (prefh <> nil) and (err <> noErr) then begin
			SetHandleSize(prefh, 0);
		end; (* if *)
		ICRFindPrefHandle := err;
	end; (* ICRFindPrefHandle *)

	function ICRSetPrefHandle (var inst: ICRRecord; key: StringPtr; attr: ICAttr; prefh: Handle): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			s: SignedByte;
	begin
		err := noErr;
		if prefh <> nil then begin
			if prefh^ = nil then begin
				err := paramErr;
			end; (* if *)
			if err = noErr then begin
				s := HGetState(prefh);
				HLock(prefh);
				err := ICSetPref(ICInstance(@inst), key^, attr, prefh^, GetHandleSize(prefh));
				HSetState(prefh, s);
			end; (* if *)
		end else begin
			err := ICSetPref(ICInstance(@inst), key^, attr, nil, 0);
		end; (* if *)
		ICRSetPrefHandle := err;
	end; (* ICRSetPrefHandle *)

	(* ***** Indexing Preferences ***** *)

	function ICRCountPref (var inst: ICRRecord; var count: longint): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			currentRefNum: integer;
	begin
		err := CheckInside(inst);
		if err = noErr then begin
			if inst.configFileRefNum = 0 then begin
				(* This is the special case where a missing file was opened
					for read-only access.  Obviously such a file has no
					preferences in it, so the answer is easy.
				*)
				count := 0;
			end else begin
				(* Switch the resource files then count the preference resources. *)
				currentRefNum := CurResFile;
				UseResFile(inst.configFileRefNum);
				err := ResError;
				if err = noErr then begin
					count := Count1Resources(kICPrefResourceType);
					err := ResError;
					UseResFile(currentRefNum);
				end; (* if *)
			end; (* if *)
		end; (* if *)

		(* Clean up. *)
		if err <> noErr then begin
			count := 0;
		end; (* if *)
		ICRCountPref := err;
	end; (* ICRCountPref *)

	function ICRGetIndPref (var inst: ICRRecord; n: longint; var key: Str255): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			currentRefNum: integer;
			prefh: Handle;
			junkResourceID: integer;
			junkResourceType: ResType;
	begin
		(* Prepare for failure. *)
		prefh := nil;

		(* Check parameters. *)
		err := CheckInside(inst);
		if (err = noErr) and (n < 1) then begin
			err := paramErr;						(* Bogus index parameter. *)
		end; (* if *)
		
		(* Do the work. *)
		if err = noErr then begin
			if inst.configFileRefNum = 0 then begin
				(* This is the special case where a missing file was opened
					for read-only access.  Obviously such a file has no
					preferences in it, so the answer is easy.
				*)
				err := icPrefNotFoundErr;
			end else begin
				(* Switch resource files, get the resource, then ask for it's name. *)
				currentRefNum := CurResFile;
				UseResFile(inst.configFileRefNum);
				err := ResError;
				if err = noErr then begin

					(* SetResLoad to false to avoid having the Resource Manager
						actually bring the preference data into RAM.  This reduces
						the footprint on the client application.
					*)
					SetResLoad(false);
					prefh := Get1IndResource(kICPrefResourceType, n);
					SetResLoad(true);
					if prefh = nil then begin
						err := icPrefNotFoundErr;
					end else begin
						GetResInfo(prefh, junkResourceID, junkResourceType, key);
						err := ResError;
					end; (* if *)
					UseResFile(currentRefNum);
				end; (* if *)
			end; (* if *)
		end; (* if *)

		(* Clean up. *)
		if prefh <> nil then begin
			ReleaseResource(prefh);
		end; (* if *)
		ICRGetIndPref := err;
	end; (* ICRGetIndPref *)

	(* ***** Deleting Preferences ***** *)

	function ICRDeletePref (var inst: ICRRecord; key: StringPtr): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			copiedKey: Str255;
			prefH: Handle;
			currentRefNum: integer;
	begin
		(* Check parameters. *)
		copiedKey := '';
		err := CheckInside(inst);
		if (err = noErr) and (key <> nil) then begin
			BlockMoveData(key, @copiedKey, ord(key^[0]) + 1);
		end; (* if *)
		if (err = noErr) and (copiedKey = '') then begin
			(* key was nil, or key was empty. *)
			err := paramErr;
		end; (* if *)
		
		(* Do the work. *)
		if err = noErr then begin
			if inst.configFileRefNum = 0 then begin
				(* This is the special case where a missing file was opened
					for read-only access.  Obviously such a file has no
					preferences in it, so the answer is easy.
				*)
				err := icPrefNotFoundErr;
			end else begin
				(* Switch resource files, get the resource, then remove it. *)
				currentRefNum := CurResFile;
				UseResFile(inst.configFileRefNum);
				err := ResError;
				if err = noErr then begin

					(* SetResLoad to false to avoid having the Resource Manager
						actually bring the preference data into RAM.  This reduces
						the footprint on the client application.
					*)
					SetResLoad(false);
					prefH := Get1NamedResource(kICPrefResourceType, copiedKey);
					err := ResError;
					SetResLoad(true);
					if (prefH = nil) or (err = resNotFound) then begin
						(* If the preference is missing, we want to return the
							correct IC error code.
						*)
						err := icPrefNotFoundErr;
					end; (* if *)

					(* Zap the preference. *)
					if err = noErr then begin
						RemoveResource(prefH);
						err := ResError;
					end; (* if *)
					UseResFile(currentRefNum);
				end; (* if *)
			end; (* if *)
		end; (* if *)
		ICRDeletePref := err;
	end; (* ICRDeletePref *)

	(* ***** User Interface Stuff ***** *)

	function ICREditPreferences (var inst: ICRRecord; key: StringPtr): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			copiedKey: Str255;
	begin
		(* Check parameters. *)
		copiedKey := '';
		err := noErr;
		if not inst.haveConfigFile then begin
			err := bdNamErr;
		end; (* if *)
		if (err = noErr) and (key <> nil) then begin
			BlockMoveData(key, @copiedKey, ord(key^[0]) + 1);
		end; (* if *)

		(* Pass the info along to ICUEditPreferences which does all the work. *)
		if err = noErr then begin
			err := ICUEditPreferences(copiedKey, inst.configFileSpec);
		end; (* if *)
		ICREditPreferences := err;
	end; (* ICREditPreferences *)

	(* ***** URL Handling Utilities ***** *)

	(* URL Parsing Algorithm

		The following is a short summary of the URL parsing algorithm:
		
		1.		If there is a selection skip to Step 4.
		2.		Expand selection in both directions to end of word (never skip
				a bracket though).
		3.		If either end has a bracket then expand other end to search for 
				a matching bracket.
		4.		Strip trailing and leading whitespace.
		5. 		Strip whitespace CR whitespace.
		6. 		Take off the outside brackets or ...
		7. 		... strip trailing "." unless there was originally a selection.
		8. 		Remove leading "URL:"
		9. 		Extract scheme by looking forwards for ":"
		10. 	If no scheme then either prepend "hint:" or just "hint" depending
				on whether hint already has a ":".
	*)

	(* This array is used to access the characters in the URL. *)
	const
		kVeryBig = 100000;
	type
		DataArray = packed array[0..kVeryBig] of char;
		DataPtr = ^DataArray;
		DataHandle = ^DataPtr;

	(* These sets represent the various types of characters in a URL. *)

	const
		kForwardExpandTerminateSet = [' ', '<', '[', '(', '{', '"', '''', '', '', chr(9), chr(13)];
		(* These characters will terminate an expansion (ie Step 2 in the algorithm
			above) in the forward direction (ie towards the front of the URL).
		*)

		kForwardBracketSet = ['<', '[', '(', '{', '"', '''', '', ''];
		(* This set contains all the characters that are considered forward
			brackets (see Step 2 and 3 in the algorithm above).
		*)

		kBackwardExpandTerminateSet = [' ', '>', ']', ')', ')', '"', '''', '', '', chr(9), chr(13)];
		(* These characters will terminate an expansion (ie Step 2 in the algorithm
			above) in the backward direction (ie towards the end of the URL).
		*)
		
		kBackwardBracketSet = ['>', ']', ')', '}', '"', '''', '', ''];
		(* This set contains all the characters that are considered backward
			brackets (see Step 2 and 3 in the algorithm above).
		*)
		
		kWhitespaceSet = [' ', chr(9)];
		kWhitespaceReturnSet = [' ', chr(9), chr(13)];
		(* These characters are considered whitespace by Steps 4 and
			5 of the algorithm above.
		*)

	function MatchingBracket (ch: char): char;
		(* Returns the matching bracket for the input character,
			eg MatchingBracket('{') = '}' and MatchingBracket('}') = '{'.
			Despite the size of the source code, this generates
			a surprisingly small executable.
		*)
	begin
		case ch of
			'<': 
				MatchingBracket := '>';
			'>': 
				MatchingBracket := '<';
			'[': 
				MatchingBracket := ']';
			']': 
				MatchingBracket := '[';
			'(': 
				MatchingBracket := ')';
			')': 
				MatchingBracket := '(';
			'{': 
				MatchingBracket := '}';
			'}': 
				MatchingBracket := '{';
			'': 
				MatchingBracket := '';
			'': 
				MatchingBracket := '';
			'': 
				MatchingBracket := '';
			'': 
				MatchingBracket := '';
			otherwise
				MatchingBracket := ch;
		end; (* case *)
	end; (* MatchingBracket *)

	function ExpandSelection (dataP: DataPtr; len: longint; var selStart, selEnd: longint): ICError;
		(* Expands the selection denoted by selStart and selEnd in the URL
			text denoted by dataP and len.  This implemenst Step 2 and 3 of the
			algorithm given above.
		*)
		var
			err: ICError;
			found: Boolean;
			searchChar: char;
	begin
		err := noErr;

		(* Expand leading selection looking for word break. If we hit a bracket,
			stop, but also include the bracket in the selection.
		*)
		while (selStart > 0) & not (dataP^[selStart - 1] in kForwardExpandTerminateSet) do begin
			selStart := selStart - 1;
		end; (* while *)
		if (selStart > 0) & (dataP^[selStart - 1] in kForwardBracketSet) then begin
			selStart := selStart - 1;
		end; (* if *)
		
		(* Expand the trailing selection looking for work break.  If we hit a bracket,
			stop, but also include the bracket in the selection.
		*)
		while (selEnd < len) & not (dataP^[selEnd] in kBackwardExpandTerminateSet) do begin
			selEnd := selEnd + 1;
		end; (* while *)
		if (selEnd < len) & (dataP^[selEnd] in kBackwardBracketSet) then begin
			selEnd := selEnd + 1;
		end; (* if *)

		(* If first character was a bracket, expand the trailing selection
			to meet the matching bracket.
		*)
		if dataP^[selStart] in kForwardBracketSet then begin
			searchChar := MatchingBracket(dataP^[selStart]);
			found := false;
			while not found and (selEnd - 1 < len) do begin
				found := (dataP^[selEnd - 1] = searchChar);
				if not found then begin
					selEnd := selEnd + 1;
				end; (* if *)
			end; (* while *)
			if not found then begin
				err := icNoURLErr;
			end; (* if *)
		end; (* if *)

		(* If last character was a bracket, expand leading selection to meet
			matching bracket.
		*)
		if (err = noErr) & (selEnd > 0) & (dataP^[selEnd - 1] in kBackwardBracketSet) then begin
			searchChar := MatchingBracket(dataP^[selEnd - 1]);
			found := (dataP^[selStart] = searchChar);
			while not found and (selStart >= 0) do begin
				found := (dataP^[selStart] = searchChar);
				if not found then begin
					selStart := selStart - 1;
				end; (* if *)
			end; (* if *)
			if not found then begin
				err := icNoURLErr;
			end; (* if *)
		end; (* if *)

		ExpandSelection := err;
	end; (* ExpandSelection *)

	function ShrinkSelection (dataP: DataPtr; len: longint; var selStart, selEnd: longint): ICError;
		(* Shrink the leading and trailing edges of the selection denoted by
			selStart and selEnd in the URL text denoted by dataP and len.
			This implemenst Step 4 of the algorithm given above.
		*)
	begin
		(*Strip leading whitespace. *)
		while (selStart < len) & (dataP^[selStart] in kWhitespaceSet) do begin
			selStart := selStart + 1;
		end; (* while *)
		(* Strip trailing whitespace. *)
		while (selEnd > 0) & (dataP^[selEnd - 1] in kWhitespaceSet) do begin
			selEnd := selEnd - 1;
		end; (* while *)
		ShrinkSelection := noErr;
	end; (* ShrinkSelection *)

	function StripReturns (urlH: DataHandle): ICError;
		(* Removes any sequence of <whitespace> <cr> <whitespace> from urlH.
			This is Step 5 of the algorithm described above.
		*)
		var
			err: ICError;
			sourceSize: longint;
			sourceIndex: longint;
			destIndex: longint;
	begin
		sourceSize := GetHandleSize(Handle(urlH));
		sourceIndex := 0;
		destIndex := 0;

		(* Skip down the handle, copying the character at sourceIndex to
			destIndex except when meeting CR where we do special things.
		*)
		while sourceIndex < sourceSize do begin

			if urlH^^[sourceIndex] = chr(13) then begin
				(* We've hit a CR.  Move destIndex back to point to previous
					non-whitespace character.  Then move sourceIndex forwards
					to next non-whitespace, skipping both normal whitespace
					and CRs.
				*)
				while (destIndex > 0) & (urlH^^[destIndex - 1] in kWhitespaceSet) do begin
					destIndex := destIndex - 1;
				end; (* while *)
				while (sourceIndex < sourceSize) & (urlH^^[sourceIndex] in kWhitespaceReturnSet) do begin
					sourceIndex := sourceIndex + 1;
				end; (* while *)
			end; (* case *)

			(* The previous code could have moved sourceIndex beyond sourceSize,
				so check again before copying a bit from sourceIndex to destIndex.
			*)
			if sourceIndex < sourceSize then begin
				urlH^^[destIndex] := urlH^^[sourceIndex];
				sourceIndex := sourceIndex + 1;
				destIndex := destIndex + 1;
			end; (* if *)
		end; (* while *)

		(* Resize the handle to the number of bytes that we copied. *)
		SetHandleSize(Handle(urlH), destIndex);
		err := MemError;
		if (err = noErr) & (GetHandleSize(Handle(urlH)) = 0) then begin
			err := icNoURLErr;
		end; (* if *)

		StripReturns := err;
	end; (* StripReturns *)

	function FindScheme (urlH: Handle; scheme: StringPtr): ICError;
		(* This routine extracts the scheme part from the URL
			by search through the handle for the first ":" and then
			copying all of the characters from the start of the
			URL to just before the ":" out into the scheme string.
			If it can't find a ":", it returns an error.
			If scheme is nil, it does not copy the string out,
			so only the error return is significant.
			FindScheme implements Step 9 of the URL parsing algorithm.
		*)
		var
			err: ICError;
			tmp: Str15;
			ndx: longint;
	begin
		err := noErr;
		tmp := ':';
		ndx := Munger(Handle(urlH), 0, @tmp[1], length(tmp), nil, 0);
		if (ndx < 0) or (ndx > 255) then begin
			err := icNoURLErr;
		end; (* if *)
		if (err = noErr) and (scheme <> nil) then begin
			scheme^[0] := chr(ndx);
			BlockMoveData(urlH^, @scheme^[1], ndx);
		end; (* if *)
		FindScheme := err;
	end; (* FindScheme *)

	(* ***** URL Handling ***** *)

	function ICRParseURL (var inst: ICRRecord; hint: StringPtr; data: Ptr; len: longint; var selStart: longint; var selEnd: longint; url: Handle): ICError;
		(* See comment in ICAPI.p.  Note the sneaky use of aliases in this
			routine.  urlH and url are the same handle, it's just that urlH
			is of type DataHandle whereas url is of type Handle.  The code
			uses them interchangably to avoid a ridiculous number of type
			casts, but it can get confusing.
		*)
		var
			err: ICError;
			dataP: DataPtr;
			urlH: DataHandle;
			explicitSelection: Boolean;
			copiedHint: Str255;
			tmp: Str15;
			junklong: longint;
	begin
		{$unused inst}
		(* Setup some locals which are retyped versions of the input parameters. *)
		dataP := DataPtr(data);
		urlH := DataHandle(url);
		explicitSelection := true;
		
		(* Check the parameters. *)
		err := noErr;
		if (data = nil) |
					(url = nil) | (url^ = nil) | (len <= 0) | 
							(selStart < 0) | (selEnd < 0) | (selStart > len) | (selEnd > len) |
									(selStart > selEnd) then begin
			err := paramErr;
		end; (* if *)
		
		(* If there's not explicit selection, expand the insertion point.  This is
			Step 2 and 3 of the algorithm.
		*)
		if (err = noErr) and (selStart = selEnd) then begin
			explicitSelection := false;
			err := ExpandSelection(dataP, len, selStart, selEnd);
		end; (* if *)

		(* Remove any leading or trailing whitespace -- Step 4 of the algorithm. *)
		if err = noErr then begin
			err := ShrinkSelection(dataP, len, selStart, selEnd);
		end; (* if *)
		if (err = noErr) and (selStart >= selEnd) then begin
			(* No selection remains.  We don't have a URL. *)
			err := icNoURLErr;
		end; (* if *)

		(* Copy the selection (ie the URL) into url handle. *)
		if err = noErr then begin
			err := PtrToXHand(@dataP^[selStart], url, selEnd - selStart);
		end; (* if *)

		(* Step 5 of the algorithm is to remove any <whitespace> <cr> <whitespace>
			sequences.
		*)
		if err = noErr then begin
			err := StripReturns(urlH);
		end; (* if *)

		if err = noErr then begin
			if (urlH^^[0] in kForwardBracketSet) &
						(urlH^^[GetHandleSize(Handle(urlH)) - 1] = MatchingBracket(urlH^^[0])) then begin
				(* If the first character of the URL is a bracket and the last
					character is the matching opposite bracket, remove the brackets.
					Step 6 of the algorithm.
				*)
				SetHandleSize(url, GetHandleSize(url) - 1);
				junklong := Munger(url, 0, nil, 1, Ptr(-1), 0);	(* trim off first character *)
			end else if not explicitSelection then begin
				(* If the last character is a "." and there was no explicit selection,
					remove it.  Step 7 of the algorithm.  This is designed to handle
					the case where losers put URLs at the end of sentences and
					fail to enclose them in brackets, for example
					http://redback.cs.uwa.edu.au/Quinn/.
				*)
				if urlH^^[GetHandleSize(Handle(urlH)) - 1] = '.' then begin
					SetHandleSize(url, GetHandleSize(url) - 1);
				end; (* if *)
			end; (* if *)
			
			(* Step 8 is to trim off a leading "URL:" if there is one.  Note the use
				of HUnlock vs HSetState.  This is cool because we know that
				the handle should be unlocked becauses we created it.
			*)
			tmp := 'URL:';
			HLock(url);
			if (GetHandleSize(url) >= length(tmp)) &
						(IUMagIDString(url^, @tmp[1], length(tmp), length(tmp)) = 0) then begin
				(* Unlock urlH 'cause Munger is going to want it that way. *)
				HUnlock(url);
				(* Trim the 'URL:' characters off the front of the URL. *)
				junklong := Munger(url, 0, nil, length(tmp), Ptr(-1), 0);
			end; (* if *)
			HUnlock(url);

			(* Search for the URL scheme -- Step 9 of the algorithm. *)
			if FindScheme(url, nil) <> noErr then begin
				(* Failed to find the scheme (ie a ":" in first 256 bytes),
					prepend "hint:" to URL.
				*)
				if (hint = nil) | (hint^[0] = chr(0)) then begin
					(* Whoops, there's no hint to prepend, we so bail out. *)
					err := icNoURLErr;
				end else begin
					(* Copy the hint into a local variable (see the note about
						copiedKey at the top of this file) and put it on
						the front of the URL.  This is the Step 10 of the algorithm.
					*)
					BlockMoveData(hint, @copiedHint, ord(hint^[0]) + 1);
					if pos(':', copiedHint) = 0 then begin
						(* Only add a colon if the hint doesn't already have one. *)
						copiedHint := concat(copiedHint, ':');
					end; (* if *)
					junklong := Munger(url, 0, nil, 0, @copiedHint[1], length(copiedHint));
					err := MemError;
				end; (* if *)
			end; (* if *)
		end; (* if *)
		ICRParseURL := err;
	end; (* ICRParseURL *)

	function ICRLaunchURL (var inst: ICRRecord; hint: StringPtr; data: Ptr; len: longint; var selStart: longint; var selEnd: longint): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			urlH: Handle;
			helper: ICAppSpec;
			scheme: Str255;
			junkAttr: longint;
			size: longint;
	begin
		(* Parse the URL out of the text into a handle. *)
		urlH := NewHandle(0);
		err := MemError;
		if err = noErr then begin
			err := ICParseURL(ICInstance(@inst), hint^, data, len, selStart, selEnd, urlH);
		end; (* if *)
		
		(* Lookup the scheme associated with the URL, then find the helper
			preference for that scheme and send that application
			the GURL AppleEvent.
		*)
		if err = noErr then begin
			err := FindScheme(urlH, @scheme);
		end; (* if *)
		if err = noErr then begin
			size := sizeof(helper);
			err := ICGetPref(ICInstance(@inst), concat(kICHelper, scheme), junkAttr, @helper, size);
			if (err = icPrefNotFoundErr) &
						(EqualString(scheme, 'https', false, true) |
								EqualString(scheme, 'file', false, true)) then begin
				(* If the user has no preference for "https" or "file" URLs, automagically
					pass them to the "http" helper.
				*)
				size := sizeof(helper);
				err := ICGetPref(ICInstance(@inst), concat(kICHelper, 'http'), junkAttr, @helper, size);
			end; (* if *)
		end; (* if *)
		if err = noErr then begin
			err := ICULaunchURL(helper.fCreator, urlH);
		end; (* if *)
		
		(* Clean up. *)
		if urlH <> nil then begin
			DisposeHandle(urlH);
		end; (* if *)
		ICRLaunchURL := err;
	end; (* ICRLaunchURL *)

	(* ***** Internal Mapping Routines *****)

	const
		kMinimumICMapEntryLength = 6;
		(* An ICMapEntry must have at least 6 bytes in it, namely
			the total_length, fixed_length and version fields.
		*)

	function UnpackEntry (entries: handle; pos: longInt; var entry: ICMapEntry; var userDataSize: longInt): OSErr;
		(* This routine unpacks the ICMapEntry from position pos in the entries
			handle.  It fills out the unpacked form of the ICMapEntry
			and the userDataSize count, which is the number of bytes
			of user data in the entry.
			
			WARNING: Depends very much on the exact format of ICMapEntry!
		*)

		procedure CopyStringOut (var p: Ptr; var s: Str255);
			(* Copies a Pascal string from p to s using BlockMoveData
				to avoid alignment problems.  It also updates p to point
				to the byte after the string.
			*)
			var
				stringLength: integer;
		begin
			stringLength := band(p^, $0FF) + 1;
			BlockMoveData(p, @s, stringLength);
			p := Ptr(ord(p) + stringLength);
		end; (* CopyStringOut *)

		var
			err: OSErr;
			entryBase: Ptr;
			maxSize: longInt;
			copyCursor: Ptr;
	begin
		(* Check parameters.  Because this internal routine is the first
			'point-of-contact' that many of the IC routines have to parameters
			passed in by the client, we do rigorous parameter checking here
			to save doing it in all of the IC routines that call here.
		*)
		err := noErr;
		if (entries = nil) | (entries^ = nil) |
					(pos < 0) | (pos > GetHandleSize(entries) - kMinimumICMapEntryLength) then begin
			err := paramErr;
		end; (* if *)

		(* Get the first 6 bytes out of the entry and do a rough check for
			their validity.  This makes sure we don't go off into the weeds
			when passed a completely bogus entries handle.
		*)
		if err = noErr then begin
			entryBase := Ptr(ord(entries^) + pos);
			maxSize := GetHandleSize(entries);
			BlockMoveData(entryBase, @entry, kMinimumICMapEntryLength);
			if (entry.fixed_length <> ICmap_fixed_length) |
						(entry.fixed_length > entry.total_length) | 
								(entry.total_length > maxSize) then begin
				(* WTF is this error code!  See what happens when you
					let PNL write the code!
				*)
				err := badExtResource;
			end; (* if *)
		end; (* if *)
		
		(* All is good.  Copy the data out of the entries handle
			into the ICMapEntry, unpacking the packed Pascal
			strings as we go.  Use copyCursor to keep track
			of the current position into the variable length
			data as we copy the data out.  At the end, 
			use the difference between copyCursor and entryBase
			to calculate the total bytes that we know about
			in the entry and hence the number of user bytes.
		*)
		if err = noErr then begin
			BlockMoveData(entryBase, @entry, entry.fixed_length);
			copyCursor := Ptr(ord(entryBase) + entry.fixed_length);
			CopyStringOut(copyCursor, entry.extension);
			CopyStringOut(copyCursor, entry.creator_app_name);
			CopyStringOut(copyCursor, entry.post_app_name);
			CopyStringOut(copyCursor, entry.MIME_type);
			CopyStringOut(copyCursor, entry.entry_name);
			userDataSize := entry.total_length - (ord(copyCursor) - ord(entryBase));
		end; (* if *)
		UnpackEntry := err;
	end; (* UnpackEntry *)

	procedure PackEntry (var entry: ICMapEntry; p: Ptr; userDataSize: longInt);
		(* Pack the ICMapEntry and copy it into the address pointed to by p.
			When setting up the entry.total_length field, leave enough
			space for userDataSize bytes of user data.
		*)

		procedure CopyStringIn (var s: Str255);
			(* Copy the string s into the client buffer pointed to by
				p in a packed fashion, using BlockMoveData to avoid
				alignment problems.  Update entry.total_length
				to ensure that the next call to this routine will
				copy the string packed immediately after this string.
			*)
		begin
			BlockMoveData(@s, Ptr(ord(p) + entry.total_length), length(s) + 1);
			entry.total_length := entry.total_length + length(s) + 1;
		end; (* CopyStringIn *)

	begin
		entry.version := 0;
		entry.fixed_length := ord(@entry.extension) - ord(@entry);

		(* Before copying the strings in, set up entry.total_length
			to make sure that CopyStringIn starts copying to the right
			place.
		*)
		entry.total_length := entry.fixed_length;
		CopyStringIn(entry.extension);
		CopyStringIn(entry.creator_app_name);
		CopyStringIn(entry.post_app_name);
		CopyStringIn(entry.MIME_type);
		CopyStringIn(entry.entry_name);
		
		(* Now that we've finished copying the strings in, update
			entry.total_length to include the number of bytes of user data.
			Then copy the fixed fields of the entry into the user buffer.
		*)
		entry.total_length := entry.total_length + userDataSize;
		BlockMoveData(@entry, p, entry.fixed_length);
	end; (* PackEntry *)

	function FastGetEntry (entries: Handle; pos: longint; var entry: ICMapEntry): OSErr;
		(* A faster version of UnpackEntry, which only returns the fixed part
			of the entry along with the extension string.  This is called by
			clients which need only this information and need to go quickly.

			WARNING: Depends very much on the exact format of ICMapEntry!
		*)
		var
			err: OSErr;
			entryBase: Ptr;
			maxSize: longInt;
			p: Ptr;
	begin
		(* Check parameters.  Because this internal routine is the first
			'point-of-contact' that many of the IC routines have to parameters
			passed in by the client, we do rigorous parameter checking here
			to save doing it in all of the IC routines that call here.
		*)
		err := noErr;
		if (entries = nil) | (entries^ = nil) |
					(pos < 0) | (pos > GetHandleSize(entries) - kMinimumICMapEntryLength) then begin
			err := paramErr;
		end; (* if *)
	
		(* Get the first 6 bytes out of the entry and do a rough check for
			their validity.  This makes sure we don't go off into the weeds
			when passed a completely bogus entries handle.
		*)
		if err = noErr then begin
			entryBase := (Ptr(ord(entries^) + pos));
			maxSize := GetHandleSize(entries);
			BlockMoveData(entryBase, @entry, kMinimumICMapEntryLength);
			if (entry.fixed_length <> ICmap_fixed_length) |
						(entry.fixed_length > entry.total_length) |
								(entry.total_length > maxSize) then begin
				err := badExtResource;
			end; (* if *)
		end; (* if *)

		(* All is good.  Copy the fixed data out of the entries handle
			into the ICMapEntry.  Also copy out the extension field,
			which is really easy to find and used by the clients
			of this routine.
		*)
		if err = noErr then begin
			BlockMoveData(entryBase, @entry, entry.fixed_length);
			p := Ptr(ord(entryBase) + entry.fixed_length);
			BlockMoveData(p, @entry.extension, band(p^, $0FF) + 1);
		end; (* if *)
		FastGetEntry := err;
	end; (* FastGetEntry *)

	function GetShort (p: Ptr): integer;
		(* Gets a short integer out of an arbitrary location,
			without worrying about alignment restrictions.
			This would be a good candidate for inlining (-:
		*)
	begin
		GetShort := band(p^, $0FF) * 256 + band(Ptr(ord(p) + 1)^, $0FF);
	end; (* if *)

	function UpCase (ch: char): char;
		(* Upper case an ASCII character the quick and dirty way. *)
		inline
			$301F,						(* 				move.w		(sp)+,d0 						*)
			$0C00, $0061,			(* 				cmpi.b		#'a',d0							*)
			$6500, $000E,			(* 				bcs			done				; d0 < 'a'	*)
			$0C00, $007B,			(* 				cmpi.b		#'{',d0							*)
			$6400, $0006,			(* 				bcc			done				; d0 >= '{'	*)
			$0400, $0020,			(* 				subi.b		#$20,d0						*)
			$3E80;						(* done:		move.w		d0,(a7)							*)

	function IsExtensionVar (var name, extension: Str255): Boolean;
		(* Returns true if name has a file extension that matches extension. *)
		var
			nameIndex : integer;
			extensionIndex: integer;
	begin
		IsExtensionVar := false;
		if length(name) >= length(extension) then begin
			nameIndex := length(name) - length(extension) + 1;
			extensionIndex := 1;
			while extensionIndex <= length(extension) do begin
				if UpCase(name[nameIndex]) <> UpCase(extension[extensionIndex]) then begin
					leave;
				end; (* if *)
				nameIndex := nameIndex + 1;
				extensionIndex := extensionIndex + 1;
			end; (* while *)
			IsExtensionVar := (extensionIndex > length(extension));
		end; (* if *)
	end; (* IsExtensionVar *)

	(* ***** Low Level Mapping Routines ***** *)

	function ICRCountMapEntries (var inst: ICRRecord; entries: Handle; var count: longint): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			entryBase: Ptr;
			pos: longint;
			entriesHandleSize : longint;
			entrySize: integer;
	begin
		(* Check parameters. *)
		{$unused inst}
		err := noErr;
		if (entries = nil) | (entries^ = nil) then begin
			err := paramErr;
		end; (* if *)
		
		(* Starting at the beginning of the entries handle, extract the total_length
			field to skip to the next entry, counting the entries as we go, until
			we run out of entries in the handle.
		*)
		if err = noErr then begin
			entryBase := entries^;
			pos := 0;
			count := 0;
			entriesHandleSize := GetHandleSize(entries);
			while pos < entriesHandleSize do begin
				entrySize := GetShort(entryBase);
				pos := pos + entrySize;
				entryBase := Ptr(ord(entryBase) + entrySize);
				count := count + 1;
			end; (* while *)
		end; (* if *)
		ICRCountMapEntries := err;
	end; (* ICRCountMapEntries *)

	function ICRGetIndMapEntry (var inst: ICRRecord; entries: handle; ndx: longint; var pos: longint; var entry: ICMapEntry): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			entryBase: Ptr;
			entrySize: integer;
			entriesHandleSize : longint;
	begin
		(* Check parameters. *)
		err := noErr;
		if (entries = nil) | (entries^ = nil) | (ndx < 0) then begin
			err := paramErr;
		end; (* if *)
		
		(* Skip through the handle (in a manner similar to that
			used by CountMapEntries) looking for the ndx'th
			entry.  When we find it, call GetMapEntry to extract it.
		*)
		if err = noErr then begin
			entryBase := entries^;
			pos := 0;
			entriesHandleSize := GetHandleSize(entries);
			while (ndx > 1) & (pos < entriesHandleSize) do begin
				entrySize := GetShort(entryBase);
				pos := pos + entrySize;
				entryBase := Ptr(ord(entryBase) + entrySize);
				ndx := ndx - 1;
			end; (* while *)
			err := ICGetMapEntry(ICInstance(@inst), entries, pos, entry);
		end; (* if *)
		ICRGetIndMapEntry := err;
	end; (* ICRGetIndMapEntry *)

	function ICRGetMapEntry (var inst: ICRRecord; entries: handle; pos: longInt; var entry: ICMapEntry): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			junkUserDataSize: longInt;
	begin
		(* Check parameters. *)
		{$unused inst}
		err := noErr;
		if (entries = nil) | (entries^ = nil) |
					(pos < 0) | (pos >= GetHandleSize(entries)) then begin
			err := paramErr;
		end; (* if *)

		(* Simply call UnpackEntry to unpack the entry for the client. *)
		if err = noErr then begin
			err := UnpackEntry(entries, pos, entry, junkUserDataSize);
		end; (* if *)
		ICRGetMapEntry := err;
	end; (* ICRGetMapEntry *)

	function ICRSetMapEntry (var inst: ICRRecord; entries: handle; pos: longInt; var entry: ICMapEntry): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			oldEntry: ICMapEntry;
			oldEntryLength: longInt;
			userDataSize: longInt;
			packedNewEntry: ICMapEntry;
			junkLong: longInt;
	begin
		(* Check parameters. *)
		{$unused inst}
		err := noErr;
		if (entries = nil) | (entries^ = nil) | (pos < 0) | (pos >= GetHandleSize(entries)) then begin
			err := paramErr;
		end; (* if *)

		(* First we have to unpack the entry to find out the size of the user data. *)
		if err = noErr then begin
			err := UnpackEntry(entries, pos, oldEntry, userDataSize);
		end; (* if *)

		(* Now we pack the new entry into packedNewEntry, and set it up
			so that it indicates there is userDataSize bytes of user data.
		*)
		if err = noErr then begin
			PackEntry(entry, @packedNewEntry, userDataSize);
			oldEntryLength := oldEntry.total_length - userDataSize;
			
			(* If there are less than 8 bytes of user data, remove it.  This
				is basically a hack to remove alignment bytes that the previous
				version of IC used to generate.  It works because we've now
				defined a structure to the user bytes such that the minimum
				size of the user bytes is 8.
			*)
			if userDataSize < 8 then begin
				oldEntryLength := oldEntry.total_length;
				packedNewEntry.total_length := packedNewEntry.total_length - userDataSize;
				userDataSize := 0;
			end; (* if *)
			
			(* Finally call Munger to delete oldEntryLength worth of bytes
				and replace them with the new entry.  Note that we don't
				delete the user bytes (because oldEntryLength doesn't
				contain the user bytes) and hence they remain in the
				handle when we insert the new entry.
			*)
			junkLong := Munger(entries, pos,
											nil, oldEntryLength,
											@packedNewEntry, packedNewEntry.total_length - userDataSize);
			err := MemError;
 		end; (* if *)
		ICRSetMapEntry := err;
	end; (* ICRSetMapEntry *)

	function ICRDeleteMapEntry (var inst: ICRRecord; entries: handle; pos: longint): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			entry: ICMapEntry;
			junkUserDataSize: longInt;
			junk: longint;
	begin
		(* Check parameters. *)
		{$unused inst}
		err := noErr;
		if (entries = nil) | (entries^ = nil) |
					(pos < 0) | (pos >= GetHandleSize(entries)) then begin
			err := paramErr;
		end; (* if *)

		(* Unpack the entry solely to get the total_length. *)
		if err = noErr then begin
			err := UnpackEntry(entries, pos, entry, junkUserDataSize);
		end; (* if *)
		
		(* Now delete that many bytes from the pos position in the handle. *)
		if err = noErr then begin
			junk := Munger(entries, pos, nil, entry.total_length, Ptr(-1), 0);
			err := MemError;
		end; (* if *)
		ICRDeleteMapEntry := err;
	end; (* ICRDeleteMapEntry *)

	function ICRAddMapEntry (var inst: ICRRecord; entries: handle; var entry: ICMapEntry): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			newEntry: ICMapEntry;
	begin
		(* Check parameters. *)
		{$unused inst}
		err := noErr;
		if (entries = nil) | (entries^ = nil) then begin
			err := paramErr;
		end; (* if *)

		(* Pack the passed in entry into newEntry, and then append it to the handle. *)
		if err = noErr then begin
			PackEntry(entry, @newEntry, 0);
			err := PtrAndHand(@newEntry, entries, entry.total_length);
		end; (* if *)
		ICRAddMapEntry := err;
	end; (* ICRAddMapEntry *)

	(* ***** Mid Level Mapping Routines ***** *)

	function ICRMapEntriesFilename (var inst: ICRRecord; entries: Handle; filename: StringPtr; var entry: ICMapEntry): ICError;
		(* See comment in ICAPI.p. *)
		(* implementation lifted directly from Space Aliens *)
		var
			err: ICError;
			copiedFilename: Str255;
			posIndex: longint;
			longestLenSoFar: integer;
			posOfLongestLenSoFar: longint;
	begin
		(* Check parameters. *)
		copiedFilename := '';
		err := noErr;
		if (err = noErr) and (filename <> nil) then begin
			BlockMoveData(filename, @copiedFilename, ord(filename^[0]) + 1);
		end; (* if *)
		if (entries = nil) | (entries^ = nil) | (copiedFilename = '') then begin
			err := paramErr;
		end; (* if *)

		(* Loop through each of the entries in the handle looking for the
			longest match.
		*)
		if err = noErr then begin
			longestLenSoFar := 0;
			posIndex := 0;
			while FastGetEntry(entries, posIndex, entry) = noErr do begin
				(* We've found a new longest match if:
					o not_incoming flag bit is clear
					o it's longer than the previous max
					o it's longer than the file name
					o it matches the last N chars of the filename
				*)
				if (length(entry.extension) > longestLenSoFar) & 
							not btst(entry.flags, ICmap_not_incoming_bit) & 
									IsExtensionVar(copiedFilename, entry.extension) then begin
					(* Record the new longest entry. *)
					longestLenSoFar := length(entry.extension);
					posOfLongestLenSoFar := posIndex;
				end; (* if *)
				(* Increment posIndex so that we get the next entry the next time
					around the loop.
				*)
				posIndex := posIndex + entry.total_length;
			end; (* while *)
		end; (* if *)

		(* If we never found a match, leave with an error. *)
		if (err = noErr) & (longestLenSoFar = 0) then begin
			err := icPrefNotFoundErr;
		end; (* if *)

		(* If all is OK, return the found entry to the client. *)
		if err = noErr then begin
			err := ICGetMapEntry(ICInstance(@inst), entries, posOfLongestLenSoFar, entry);
		end; (* if *)
		ICRMapEntriesFilename := err;
	end; (* ICRMapEntriesFilename *)

	function ICRMapEntriesTypeCreator (var inst: ICRRecord; entries: Handle; fType: OSType; fCreator: OSType; filename: StringPtr; var entry: ICMapEntry): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			copiedFilename: Str255;
			posIndex: longint;
			weightOfThisEntry: longint;
			bestWeightSoFar: longint;
			posOfBestWeightSoFar: longint;
	begin
		(* Check parameters. *)
		err := noErr;
		if (entries = nil) | (entries^ = nil) then begin
			err := paramErr;
		end; (* if *)

		(* Copy in the filename string to a local variable.  Note that we
			don't check it for the empty string because empty strings
			are legal for this routine.
		*)
		copiedFilename := '';
		if filename <> nil then begin
			BlockMoveData(filename, @copiedFilename, ord(filename^[0]) + 1);
		end; (* if *)
		
		(* Loop through each of the entries in the handle looking for the
			best match.  "Best match" is defined by:
			o entry must not be marked "not outgoing"
			o entry type must match fType parameter
			o entries with longer extensions that match are better
			o entries with matching creators are better

			This is implemented by a weighting mechanism.  Entries
			gain a weight of 1 if the creators match.  Entries gain
			a weight of 2 * length(extension) if the extensions match.
		*)
		if err = noErr then begin
			posIndex := 0;
			bestWeightSoFar := -1;
			while FastGetEntry(entries, posIndex, entry) = noErr do begin
				(* Entries marked "not outgoing" are not eligible. *)
				if not btst(entry.flags, ICmap_not_outgoing_bit) then begin
					if entry.file_type = fType then begin
						weightOfThisEntry := ord(entry.file_creator = fCreator);
						if IsExtensionVar(copiedFilename, entry.extension) then begin
							weightOfThisEntry := weightOfThisEntry + 2 * length(entry.extension);
						end; (* if *)
						if weightOfThisEntry > bestWeightSoFar then begin
							(* Record the new "best match" entry. *)
							posOfBestWeightSoFar := posIndex;
							bestWeightSoFar := weightOfThisEntry;
						end; (* if *)
					end; (* if *)
				end; (* if *)
				(* Increment posIndex so that we get the next entry the next time
					around the loop.
				*)
				posIndex := posIndex + entry.total_length;
			end; (* while *)

			(* If we never found a match, leave with an error, otherwise
				return the found entry to the client.
			*)
			if bestWeightSoFar = -1 then begin
				err := icPrefNotFoundErr;
			end else begin
				err := ICGetMapEntry(ICInstance(@inst), entries, posOfBestWeightSoFar, entry);
			end; (* if *)
		end; (* if *)
		ICRMapEntriesTypeCreator := err;
	end; (* ICRMapEntriesTypeCreator *)

	(* ***** High Level Mapping Routines ***** *)

	function ICRMapFilename (var inst: ICRRecord; filename: StringPtr; var entry: ICMapEntry): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			entries: Handle;
			junkAttr: ICAttr;
	begin
		err := ICGetPrefHandle(ICInstance(@inst), kICMapping, junkAttr, entries);
		if err = noErr then begin
			err := ICMapEntriesFilename(ICInstance(@inst), entries, filename^, entry);
			DisposeHandle(entries);
		end; (* if *)
		ICRMapFilename := err;
	end; (* ICRMapFilename *)

	function ICRMapTypeCreator (var inst: ICRRecord; fType: OSType; fCreator: OSType; filename: StringPtr; var entry: ICMapEntry): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			entries: Handle;
			junkAttr: ICAttr;
	begin
		err := ICGetPrefHandle(ICInstance(@inst), kICMapping, junkAttr, entries);
		if err = noErr then begin
			err := ICMapEntriesTypeCreator(ICInstance(@inst), entries, fType, fCreator, filename^, entry);
			DisposeHandle(entries);
		end; (* if *)
		ICRMapTypeCreator := err;
	end; (* ICRMapTypeCreator *)

end. (* ICLinkIn *)