unit ICSwitchGlue;

(* Start Header *)

(* File:         ICSwitchGlue.p (impl)
 * Generated by: 1.0d5
 * For:          IC 1.4
 * On:           Saturday, 22 February 1997, 17:54:20
 * 
 * This file is part of the Internet Configuration system and
 * is placed in the public domain for the benefit of all.
 *)

(* End Header *)

(* This file contains the "switch glue", which is code which is harded linked
	to the application.  This code implements the ICAPI itself.  Each entry point
	determines whether the IC component is installed.  If it is, it routes the call
	to the component, otherwise to the link-in implementation.  Hence the name
	switch glue.
	
	The vast bulk of this code is generated automatically.  The only interesting
	routines are Start and Stop, which are necessarily special cases.
*)

interface

	uses
		Types, 
		Files, 
		QuickDraw, 
		AppleTalk, 
		Aliases, 

		ICTypes, 
		ICKeys;

(* Start ICSG.p *)

  (* ***** Starting Up and Shutting Down ***** *)
  
  function ICStart(var inst : ICInstance; creator : OSType) : ICError;
  
  function ICStop(inst : ICInstance) : ICError;
  
  (* ***** Specifying a Configuration ***** *)
  
  function ICFindConfigFile(inst : ICInstance; count : integer; folders : ICDirSpecArrayPtr) : ICError;
  
  function ICFindUserConfigFile(inst : ICInstance; var where : ICDirSpec) : ICError;
  
  function ICGeneralFindConfigFile(inst : ICInstance; searchPrefs : Boolean; canCreate : Boolean; count : integer; folders : ICDirSpecArrayPtr) : ICError;
  
  function ICChooseConfig(inst : ICInstance) : ICError;
  
  function ICChooseNewConfig(inst : ICInstance) : ICError;
  
  function ICGetConfigName(inst : ICInstance; longname : Boolean; var name : Str255) : ICError;
  
  function ICGetConfigReference(inst : ICInstance; ref : ICConfigRefHandle) : ICError;
  
  function ICSetConfigReference(inst : ICInstance; ref : ICConfigRefHandle; flags : longint) : ICError;
  
  function ICSpecifyConfigFile(inst : ICInstance; var config : FSSpec) : ICError;
  
  (* ***** Getting Information ***** *)
  
  function ICGetSeed(inst : ICInstance; var seed : longint) : ICError;
  
  function ICGetPerm(inst : ICInstance; var perm : ICPerm) : ICError;
  
  function ICDefaultFileName(inst : ICInstance; var name : Str63) : ICError;
  
  function ICGetComponentInstance(inst : ICInstance; var component_inst : univ Ptr) : ICError;
  
  (* ***** Reading and Writing Preferences ***** *)
  
  function ICBegin(inst : ICInstance; perm : ICPerm) : ICError;
  
  function ICGetPref(inst : ICInstance; key : StringPtr; var attr : ICAttr; buf : Ptr; var size : longint) : ICError;
  
  function ICSetPref(inst : ICInstance; key : StringPtr; attr : ICAttr; buf : Ptr; size : longint) : ICError;
  
  function ICFindPrefHandle(inst : ICInstance; key : StringPtr; var attr : ICAttr; prefh : Handle) : ICError;
  
  function ICGetPrefHandle(inst : ICInstance; key : StringPtr; var attr : ICAttr; var prefh : Handle) : ICError;
  
  function ICSetPrefHandle(inst : ICInstance; key : StringPtr; attr : ICAttr; prefh : Handle) : ICError;
  
  function ICCountPref(inst : ICInstance; var count : longint) : ICError;
  
  function ICGetIndPref(inst : ICInstance; n : longint; var key : Str255) : ICError;
  
  function ICDeletePref(inst : ICInstance; key : StringPtr) : ICError;
  
  function ICEnd(inst : ICInstance) : ICError;
  
  (* ***** User Interface Stuff ***** *)
  
  function ICEditPreferences(inst : ICInstance; key : StringPtr) : ICError;
  
  (* ***** URL Handling ***** *)
  
  function ICParseURL(inst : ICInstance; hint : StringPtr; data : Ptr; len : longint; var selStart : longint; var selEnd : longint; url : Handle) : ICError;
  
  function ICLaunchURL(inst : ICInstance; hint : StringPtr; data : Ptr; len : longint; var selStart : longint; var selEnd : longint) : ICError;
  
  (* ***** Mappings Routines *****
   * 
   * Routines for interrogating mappings database.
   * 
   * ----- High Level Routines -----
   *)
  
  function ICMapFilename(inst : ICInstance; filename : StringPtr; var entry : ICMapEntry) : ICError;
  
  function ICMapTypeCreator(inst : ICInstance; fType : OSType; fCreator : OSType; filename : StringPtr; var entry : ICMapEntry) : ICError;
  
  (* ----- Mid Level Routines ----- *)
  
  function ICMapEntriesFilename(inst : ICInstance; entries : Handle; filename : StringPtr; var entry : ICMapEntry) : ICError;
  
  function ICMapEntriesTypeCreator(inst : ICInstance; entries : Handle; fType : OSType; fCreator : OSType; filename : StringPtr; var entry : ICMapEntry) : ICError;
  
  (* ----- Low Level Routines ----- *)
  
  function ICCountMapEntries(inst : ICInstance; entries : Handle; var count : longint) : ICError;
  
  function ICGetIndMapEntry(inst : ICInstance; entries : Handle; ndx : longint; var pos : longint; var entry : ICMapEntry) : ICError;
  
  function ICGetMapEntry(inst : ICInstance; entries : Handle; pos : longint; var entry : ICMapEntry) : ICError;
  
  function ICSetMapEntry(inst : ICInstance; entries : Handle; pos : longint; var entry : ICMapEntry) : ICError;
  
  function ICDeleteMapEntry(inst : ICInstance; entries : Handle; pos : longint) : ICError;
  
  function ICAddMapEntry(inst : ICInstance; entries : Handle; var entry : ICMapEntry) : ICError;
  

(* End ICSG.p *)

implementation

	uses
		Components, 
		Memory, 
		Errors,

		ICCAPIGlue, 
		ICLinkIn;

	function ICStart (var inst: ICInstance; creator: OSType): ICError;
		(* The standalone glue special cases the Start routine.  Here
			we allocate the ICRRecord in the client's heap, the pointer to
			which we return to the client as an ICInstance.  We then attempt to
			open a component instance.  If that succeeds, we record
			the instance and we're done.  If it fails, we create
			a standalone instance by calling the link-in implementation
			directly.
		*)
		var
			err: ICError;
	begin
		inst := NewPtr(sizeof(ICRRecord));
		err := MemError;
		if err = noErr then begin
			err := ICCStart(ICRRecordPtr(inst)^.instance, creator);
			if err <> noErr then begin
				err := ICRStart(ICRRecordPtr(inst)^, creator);
			end; (* if *)
			if err <> noErr then begin
				DisposePtr(inst);
				inst := nil;
			end; (* if *)
		end; (* if *)
		ICStart := err;
	end; (* ICStart *)

	function ICStop (inst: ICInstance): ICError;
		(* The standalone glue also special cases the Stop routine.  The
			first part is the standard switch glue that passes the Stop
			along to either the component glue or directly to link-in 
			implementation.  Then we dispose of the ICRRecord that
			we allocated in the client's heap.
		*)
		var
			err: ICError;
	begin
		if ICRRecordPtr(inst)^.instance <> nil then begin
			err := ICCStop(ICRRecordPtr(inst)^.instance);
		end else begin
			err := ICRStop(ICRRecordPtr(inst)^);
		end; (* if *)
		DisposePtr(inst);
		ICStop := err;
	end; (* ICStop *)

	function ICGetComponentInstance (inst: ICInstance; var component_inst: univ Ptr): ICError;
		(* GetComponentInstance is another special case routine because it's
			only available to clients using the ICAPI.  The implementaion is
			rather trivial.
		*)
	begin
		component_inst := Ptr(ICRRecordPtr(inst)^.instance);
		if component_inst = nil then begin
			ICGetComponentInstance := badComponentInstance;
		end else begin
			ICGetComponentInstance := noErr;
		end; (* if *)
	end; (* ICGetComponentInstance *)

(* Start ICSG.p1 *)

  function ICFindConfigFile(inst : ICInstance; count : integer; folders : ICDirSpecArrayPtr) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICFindConfigFile := ICCFindConfigFile(ICRRecordPtr(inst)^.instance, count, folders);
    end else begin
      ICFindConfigFile := ICRFindConfigFile(ICRRecordPtr(inst)^, count, folders);
    end; (* if *)
  end; (* ICFindConfigFile *)
  
  function ICFindUserConfigFile(inst : ICInstance; var where : ICDirSpec) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICFindUserConfigFile := ICCFindUserConfigFile(ICRRecordPtr(inst)^.instance, where);
    end else begin
      ICFindUserConfigFile := ICRFindUserConfigFile(ICRRecordPtr(inst)^, where);
    end; (* if *)
  end; (* ICFindUserConfigFile *)
  
  function ICGeneralFindConfigFile(inst : ICInstance; searchPrefs : Boolean; canCreate : Boolean; count : integer; folders : ICDirSpecArrayPtr) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICGeneralFindConfigFile := ICCGeneralFindConfigFile(ICRRecordPtr(inst)^.instance, searchPrefs, canCreate, count, folders);
    end else begin
      ICGeneralFindConfigFile := ICRGeneralFindConfigFile(ICRRecordPtr(inst)^, searchPrefs, canCreate, count, folders);
    end; (* if *)
  end; (* ICGeneralFindConfigFile *)
  
  function ICChooseConfig(inst : ICInstance) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICChooseConfig := ICCChooseConfig(ICRRecordPtr(inst)^.instance);
    end else begin
      ICChooseConfig := ICRChooseConfig(ICRRecordPtr(inst)^);
    end; (* if *)
  end; (* ICChooseConfig *)
  
  function ICChooseNewConfig(inst : ICInstance) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICChooseNewConfig := ICCChooseNewConfig(ICRRecordPtr(inst)^.instance);
    end else begin
      ICChooseNewConfig := ICRChooseNewConfig(ICRRecordPtr(inst)^);
    end; (* if *)
  end; (* ICChooseNewConfig *)
  
  function ICGetConfigName(inst : ICInstance; longname : Boolean; var name : Str255) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICGetConfigName := ICCGetConfigName(ICRRecordPtr(inst)^.instance, longname, name);
    end else begin
      ICGetConfigName := ICRGetConfigName(ICRRecordPtr(inst)^, longname, name);
    end; (* if *)
  end; (* ICGetConfigName *)
  
  function ICGetConfigReference(inst : ICInstance; ref : ICConfigRefHandle) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICGetConfigReference := ICCGetConfigReference(ICRRecordPtr(inst)^.instance, ref);
    end else begin
      ICGetConfigReference := ICRGetConfigReference(ICRRecordPtr(inst)^, ref);
    end; (* if *)
  end; (* ICGetConfigReference *)
  
  function ICSetConfigReference(inst : ICInstance; ref : ICConfigRefHandle; flags : longint) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICSetConfigReference := ICCSetConfigReference(ICRRecordPtr(inst)^.instance, ref, flags);
    end else begin
      ICSetConfigReference := ICRSetConfigReference(ICRRecordPtr(inst)^, ref, flags);
    end; (* if *)
  end; (* ICSetConfigReference *)
  
  function ICSpecifyConfigFile(inst : ICInstance; var config : FSSpec) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICSpecifyConfigFile := ICCSpecifyConfigFile(ICRRecordPtr(inst)^.instance, config);
    end else begin
      ICSpecifyConfigFile := ICRSpecifyConfigFile(ICRRecordPtr(inst)^, config);
    end; (* if *)
  end; (* ICSpecifyConfigFile *)
  
  function ICGetSeed(inst : ICInstance; var seed : longint) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICGetSeed := ICCGetSeed(ICRRecordPtr(inst)^.instance, seed);
    end else begin
      ICGetSeed := ICRGetSeed(ICRRecordPtr(inst)^, seed);
    end; (* if *)
  end; (* ICGetSeed *)
  
  function ICGetPerm(inst : ICInstance; var perm : ICPerm) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICGetPerm := ICCGetPerm(ICRRecordPtr(inst)^.instance, perm);
    end else begin
      ICGetPerm := ICRGetPerm(ICRRecordPtr(inst)^, perm);
    end; (* if *)
  end; (* ICGetPerm *)
  
  function ICDefaultFileName(inst : ICInstance; var name : Str63) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICDefaultFileName := ICCDefaultFileName(ICRRecordPtr(inst)^.instance, name);
    end else begin
      ICDefaultFileName := ICRDefaultFileName(ICRRecordPtr(inst)^, name);
    end; (* if *)
  end; (* ICDefaultFileName *)
  
  function ICBegin(inst : ICInstance; perm : ICPerm) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICBegin := ICCBegin(ICRRecordPtr(inst)^.instance, perm);
    end else begin
      ICBegin := ICRBegin(ICRRecordPtr(inst)^, perm);
    end; (* if *)
  end; (* ICBegin *)
  
  function ICGetPref(inst : ICInstance; key : StringPtr; var attr : ICAttr; buf : Ptr; var size : longint) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICGetPref := ICCGetPref(ICRRecordPtr(inst)^.instance, key^, attr, buf, size);
    end else begin
      ICGetPref := ICRGetPref(ICRRecordPtr(inst)^, key, attr, buf, size);
    end; (* if *)
  end; (* ICGetPref *)
  
  function ICSetPref(inst : ICInstance; key : StringPtr; attr : ICAttr; buf : Ptr; size : longint) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICSetPref := ICCSetPref(ICRRecordPtr(inst)^.instance, key^, attr, buf, size);
    end else begin
      ICSetPref := ICRSetPref(ICRRecordPtr(inst)^, key, attr, buf, size);
    end; (* if *)
  end; (* ICSetPref *)
  
  function ICFindPrefHandle(inst : ICInstance; key : StringPtr; var attr : ICAttr; prefh : Handle) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICFindPrefHandle := ICCFindPrefHandle(ICRRecordPtr(inst)^.instance, key^, attr, prefh);
    end else begin
      ICFindPrefHandle := ICRFindPrefHandle(ICRRecordPtr(inst)^, key, attr, prefh);
    end; (* if *)
  end; (* ICFindPrefHandle *)
  
  function ICGetPrefHandle(inst : ICInstance; key : StringPtr; var attr : ICAttr; var prefh : Handle) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICGetPrefHandle := ICCGetPrefHandle(ICRRecordPtr(inst)^.instance, key^, attr, prefh);
    end else begin
      ICGetPrefHandle := ICRGetPrefHandle(ICRRecordPtr(inst)^, key, attr, prefh);
    end; (* if *)
  end; (* ICGetPrefHandle *)
  
  function ICSetPrefHandle(inst : ICInstance; key : StringPtr; attr : ICAttr; prefh : Handle) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICSetPrefHandle := ICCSetPrefHandle(ICRRecordPtr(inst)^.instance, key^, attr, prefh);
    end else begin
      ICSetPrefHandle := ICRSetPrefHandle(ICRRecordPtr(inst)^, key, attr, prefh);
    end; (* if *)
  end; (* ICSetPrefHandle *)
  
  function ICCountPref(inst : ICInstance; var count : longint) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICCountPref := ICCCountPref(ICRRecordPtr(inst)^.instance, count);
    end else begin
      ICCountPref := ICRCountPref(ICRRecordPtr(inst)^, count);
    end; (* if *)
  end; (* ICCountPref *)
  
  function ICGetIndPref(inst : ICInstance; n : longint; var key : Str255) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICGetIndPref := ICCGetIndPref(ICRRecordPtr(inst)^.instance, n, key);
    end else begin
      ICGetIndPref := ICRGetIndPref(ICRRecordPtr(inst)^, n, key);
    end; (* if *)
  end; (* ICGetIndPref *)
  
  function ICDeletePref(inst : ICInstance; key : StringPtr) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICDeletePref := ICCDeletePref(ICRRecordPtr(inst)^.instance, key^);
    end else begin
      ICDeletePref := ICRDeletePref(ICRRecordPtr(inst)^, key);
    end; (* if *)
  end; (* ICDeletePref *)
  
  function ICEnd(inst : ICInstance) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICEnd := ICCEnd(ICRRecordPtr(inst)^.instance);
    end else begin
      ICEnd := ICREnd(ICRRecordPtr(inst)^);
    end; (* if *)
  end; (* ICEnd *)
  
  function ICEditPreferences(inst : ICInstance; key : StringPtr) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICEditPreferences := ICCEditPreferences(ICRRecordPtr(inst)^.instance, key^);
    end else begin
      ICEditPreferences := ICREditPreferences(ICRRecordPtr(inst)^, key);
    end; (* if *)
  end; (* ICEditPreferences *)
  
  function ICParseURL(inst : ICInstance; hint : StringPtr; data : Ptr; len : longint; var selStart : longint; var selEnd : longint; url : Handle) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICParseURL := ICCParseURL(ICRRecordPtr(inst)^.instance, hint^, data, len, selStart, selEnd, url);
    end else begin
      ICParseURL := ICRParseURL(ICRRecordPtr(inst)^, hint, data, len, selStart, selEnd, url);
    end; (* if *)
  end; (* ICParseURL *)
  
  function ICLaunchURL(inst : ICInstance; hint : StringPtr; data : Ptr; len : longint; var selStart : longint; var selEnd : longint) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICLaunchURL := ICCLaunchURL(ICRRecordPtr(inst)^.instance, hint^, data, len, selStart, selEnd);
    end else begin
      ICLaunchURL := ICRLaunchURL(ICRRecordPtr(inst)^, hint, data, len, selStart, selEnd);
    end; (* if *)
  end; (* ICLaunchURL *)
  
  function ICMapFilename(inst : ICInstance; filename : StringPtr; var entry : ICMapEntry) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICMapFilename := ICCMapFilename(ICRRecordPtr(inst)^.instance, filename^, entry);
    end else begin
      ICMapFilename := ICRMapFilename(ICRRecordPtr(inst)^, filename, entry);
    end; (* if *)
  end; (* ICMapFilename *)
  
  function ICMapTypeCreator(inst : ICInstance; fType : OSType; fCreator : OSType; filename : StringPtr; var entry : ICMapEntry) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICMapTypeCreator := ICCMapTypeCreator(ICRRecordPtr(inst)^.instance, fType, fCreator, filename^, entry);
    end else begin
      ICMapTypeCreator := ICRMapTypeCreator(ICRRecordPtr(inst)^, fType, fCreator, filename, entry);
    end; (* if *)
  end; (* ICMapTypeCreator *)
  
  function ICMapEntriesFilename(inst : ICInstance; entries : Handle; filename : StringPtr; var entry : ICMapEntry) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICMapEntriesFilename := ICCMapEntriesFilename(ICRRecordPtr(inst)^.instance, entries, filename^, entry);
    end else begin
      ICMapEntriesFilename := ICRMapEntriesFilename(ICRRecordPtr(inst)^, entries, filename, entry);
    end; (* if *)
  end; (* ICMapEntriesFilename *)
  
  function ICMapEntriesTypeCreator(inst : ICInstance; entries : Handle; fType : OSType; fCreator : OSType; filename : StringPtr; var entry : ICMapEntry) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICMapEntriesTypeCreator := ICCMapEntriesTypeCreator(ICRRecordPtr(inst)^.instance, entries, fType, fCreator, filename^, entry);
    end else begin
      ICMapEntriesTypeCreator := ICRMapEntriesTypeCreator(ICRRecordPtr(inst)^, entries, fType, fCreator, filename, entry);
    end; (* if *)
  end; (* ICMapEntriesTypeCreator *)
  
  function ICCountMapEntries(inst : ICInstance; entries : Handle; var count : longint) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICCountMapEntries := ICCCountMapEntries(ICRRecordPtr(inst)^.instance, entries, count);
    end else begin
      ICCountMapEntries := ICRCountMapEntries(ICRRecordPtr(inst)^, entries, count);
    end; (* if *)
  end; (* ICCountMapEntries *)
  
  function ICGetIndMapEntry(inst : ICInstance; entries : Handle; ndx : longint; var pos : longint; var entry : ICMapEntry) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICGetIndMapEntry := ICCGetIndMapEntry(ICRRecordPtr(inst)^.instance, entries, ndx, pos, entry);
    end else begin
      ICGetIndMapEntry := ICRGetIndMapEntry(ICRRecordPtr(inst)^, entries, ndx, pos, entry);
    end; (* if *)
  end; (* ICGetIndMapEntry *)
  
  function ICGetMapEntry(inst : ICInstance; entries : Handle; pos : longint; var entry : ICMapEntry) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICGetMapEntry := ICCGetMapEntry(ICRRecordPtr(inst)^.instance, entries, pos, entry);
    end else begin
      ICGetMapEntry := ICRGetMapEntry(ICRRecordPtr(inst)^, entries, pos, entry);
    end; (* if *)
  end; (* ICGetMapEntry *)
  
  function ICSetMapEntry(inst : ICInstance; entries : Handle; pos : longint; var entry : ICMapEntry) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICSetMapEntry := ICCSetMapEntry(ICRRecordPtr(inst)^.instance, entries, pos, entry);
    end else begin
      ICSetMapEntry := ICRSetMapEntry(ICRRecordPtr(inst)^, entries, pos, entry);
    end; (* if *)
  end; (* ICSetMapEntry *)
  
  function ICDeleteMapEntry(inst : ICInstance; entries : Handle; pos : longint) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICDeleteMapEntry := ICCDeleteMapEntry(ICRRecordPtr(inst)^.instance, entries, pos);
    end else begin
      ICDeleteMapEntry := ICRDeleteMapEntry(ICRRecordPtr(inst)^, entries, pos);
    end; (* if *)
  end; (* ICDeleteMapEntry *)
  
  function ICAddMapEntry(inst : ICInstance; entries : Handle; var entry : ICMapEntry) : ICError;
  begin
    if ICRRecordPtr(inst)^.instance <> nil then begin
      ICAddMapEntry := ICCAddMapEntry(ICRRecordPtr(inst)^.instance, entries, entry);
    end else begin
      ICAddMapEntry := ICRAddMapEntry(ICRRecordPtr(inst)^, entries, entry);
    end; (* if *)
  end; (* ICAddMapEntry *)
  

(* End ICSG.p1 *)

end. (* ICSwitchGlue *)